import itertools
import logging
import os
from collections.abc import Iterable, Sequence
from typing import (
    Any,
    Optional,
)

import argilla as rg

from pharia_studio_sdk.connectors.argilla.argilla_client import (
    ArgillaClient,
    ArgillaEvaluation,
    Record,
    RecordData,
)


class ArgillaWrapperClient(ArgillaClient):
    def __init__(
        self,
        api_url: Optional[str] = None,
        api_key: Optional[str] = None,
        disable_warnings: bool = True,
    ) -> None:
        if disable_warnings:
            import warnings

            warnings.filterwarnings("ignore", module="argilla.*")
            # this logger is set on info for some reason
            logging.getLogger("argilla.client.feedback.dataset.local.mixins").setLevel(
                logging.WARNING
            )

        self.client = rg.Argilla(
            api_url=api_url if api_url is not None else os.getenv("ARGILLA_API_URL"),
            api_key=api_key if api_key is not None else os.getenv("ARGILLA_API_KEY"),
        )

    def create_dataset(
        self,
        workspace_name: str,
        dataset_name: str,
        fields: Sequence[rg.TextField],
        questions: Sequence[rg.QuestionType],
    ) -> str:
        """Creates and publishes a new feedback dataset in Argilla.

        Raises an error if the name exists already.

        Args:
            workspace_name: the name of the workspace the feedback dataset should be created in.
                The user executing this request must have corresponding permissions for this workspace.
            dataset_name: the name of the feedback-dataset to be created.
            fields: all fields of this dataset.
            questions: all questions for this dataset.

        Returns:
            The id of the created dataset.
        """
        settings = rg.Settings(
            fields=list(fields),
            questions=list(questions),
            allow_extra_metadata=True,
        )

        workspace = self.ensure_workspace_exists(workspace_name)

        dataset = rg.Dataset(
            name=dataset_name,
            settings=settings,
            workspace=workspace,
            client=self.client,
        )
        dataset.create()
        return str(dataset.id)

    def ensure_dataset_exists(
        self,
        workspace_name: str,
        dataset_name: str,
        fields: Sequence[rg.TextField],
        questions: Sequence[rg.QuestionType],
    ) -> str:
        """Retrieves an existing dataset or creates and publishes a new feedback dataset in Argilla.

        Args:
            workspace_name: the name of the workspace the feedback dataset should be created in.
                The user executing this request must have corresponding permissions for this workspace.
            dataset_name: the name of the feedback-dataset to be created.
            fields: all fields of this dataset.
            questions: all questions for this dataset.

        Returns:
            The id of the dataset to be retrieved .
        """
        dataset = self.client.datasets(name=dataset_name, workspace=workspace_name)
        return (
            str(dataset.id)
            if dataset
            else self.create_dataset(workspace_name, dataset_name, fields, questions)
        )

    def add_record(self, dataset_id: str, record: RecordData) -> None:
        self.add_records(dataset_id=dataset_id, records=[record])

    def add_records(self, dataset_id: str, records: Sequence[RecordData]) -> None:
        remote_dataset = self._dataset_from_id(dataset_id=dataset_id)
        if remote_dataset is None:
            raise ValueError
        argilla_records = [
            rg.Record(
                fields=dict(record.content),
                metadata={
                    **record.metadata,
                    "example_id": record.example_id,
                },
            )
            for record in records
        ]
        remote_dataset.records.log(records=argilla_records)

    def evaluations(self, dataset_id: str) -> Iterable[ArgillaEvaluation]:
        remote_dataset = self._dataset_from_id(dataset_id=dataset_id)

        status_filter = rg.Filter([("response.status", "==", "submitted")])
        query = rg.Query(filter=status_filter)

        for record in remote_dataset.records(query=query):
            metadata = record.metadata
            example_id = metadata.pop("example_id")
            yield ArgillaEvaluation(
                example_id=example_id,
                record_id="ignored",
                responses={
                    response.question_name: response.value
                    for response in record.responses
                    if response is not None
                },
                metadata=metadata,
            )

    def split_dataset(self, dataset_id: str, n_splits: int) -> None:
        """Adds a new metadata property to the dataset and assigns a split to each record.

        Deletes the property if n_splits is equal to one.

        Args:
            dataset_id: the id of the dataset
            n_splits: the number of splits to create
        """
        remote_dataset = self._dataset_from_id(dataset_id=dataset_id)

        name = "split"

        metadata_config = remote_dataset.settings.metadata[name]

        if n_splits == 1:
            if metadata_config is None:
                return
            metadata_config.delete()
            self._delete_metadata_from_records(remote_dataset, name)
            return

        if metadata_config:
            metadata_config.delete()
            self._delete_metadata_from_records(remote_dataset, name)

        self._update_record_metadata(n_splits, remote_dataset, name)

    def _update_record_metadata(
        self, n_splits: int, remote_dataset: rg.Dataset, metadata_name: str
    ) -> None:
        config = rg.IntegerMetadataProperty(
            name=metadata_name, visible_for_annotators=True, min=1, max=n_splits
        )
        remote_dataset.settings.metadata = [config]
        remote_dataset.update()
        updated_records = []
        for record, split in zip(
            remote_dataset.records, itertools.cycle(range(1, n_splits + 1))
        ):
            record.metadata[metadata_name] = split
            updated_records.append(record)

        remote_dataset.records.log(updated_records)

    def _delete_metadata_from_records(
        self, remote_dataset: rg.Dataset, metadata_name: str
    ) -> None:
        modified_records = []
        for record in remote_dataset.records:
            del record.metadata[metadata_name]
            modified_records.append(record)
        remote_dataset.records.log(modified_records)

    def ensure_workspace_exists(self, workspace_name: str) -> str:
        """Retrieves the name of an argilla workspace with specified name or creates a new workspace if necessary.

        Args:
            workspace_name: the name of the workspace to be retrieved or created.

        Returns:
            The name of an argilla workspace with the given `workspace_name`.
        """
        workspace = self.client.workspaces(name=workspace_name)
        if workspace:
            return workspace_name

        workspace = rg.Workspace(name=workspace_name, client=self.client)
        workspace.create()
        if not workspace:
            raise ValueError(
                f"Workspace with name {workspace_name} could not be created."
            )
        return str(workspace.name)

    def records(self, dataset_id: str) -> Iterable[Record]:
        remote_dataset = self._dataset_from_id(dataset_id=dataset_id)
        return (
            Record(
                id=str(record.id),
                example_id=record.metadata["example_id"],
                content=record.fields,
                metadata=record.metadata,
            )
            for record in remote_dataset.records
        )

    def _create_evaluation(
        self, dataset_id: str, record_id: str, data: dict[str, Any]
    ) -> None:
        dataset = self._dataset_from_id(dataset_id=dataset_id)
        if dataset is None:
            raise ValueError(f"Dataset with id {dataset_id} does not exist.")
        records = dataset.records

        user_id = self.client.me.id
        if user_id is None:
            raise ValueError("user_id is not a UUID")

        # argilla currently does not allow to retrieve a record by id
        # This could be optimized (in a scenario for creating multiple evaluations) by passing a dict of record_ids to the function
        # and update all the records for the given record id list.
        for record in records:
            if record.id == record_id:
                for question_name, response_value in data.items():
                    response = rg.Response(
                        question_name=question_name,
                        value=response_value,
                        status="submitted",
                        user_id=user_id,
                    )
                    record.responses.add(response)
                dataset.records.log([record])
                return

    def _delete_dataset(self, dataset_id: str) -> None:
        remote_dataset = self._dataset_from_id(dataset_id=dataset_id)
        remote_dataset.delete()

    def _delete_workspace(self, workspace_name: str) -> None:
        workspace = self.client.workspaces(name=workspace_name)
        if workspace is None:
            raise ValueError("workspace does not exist.")
        for dataset in workspace.datasets:
            dataset.delete()
        workspace.delete()

    def _dataset_from_id(self, dataset_id: str) -> rg.Dataset:
        dataset = self.client.datasets(id=dataset_id)
        if not dataset:
            raise ValueError("Dataset is not existent")
        # Fetch Settings from Dataset in order to pass questions, necessary since Argilla V2
        dataset.settings.get()
        return dataset



from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from multiprocessing import Lock as lock
from multiprocessing.synchronize import Lock
from typing import Optional, final

from pharia_inference_sdk.core import Output, Tracer
from pydantic import BaseModel

from pharia_studio_sdk.evaluation.run.domain import (
    ExampleOutput,
    FailedExampleRun,
    RunOverview,
)


class RecoveryData(BaseModel):
    run_id: str
    finished_examples: list[str] = []


class RunRepository(ABC):
    """Base run repository interface.

    Provides methods to store and load run results: :class:`RunOverview` and :class:`ExampleOutput`.
    A :class:`RunOverview` is created from and is linked (by its ID) to multiple :class:`ExampleOutput`s
    representing results of a dataset.
    """

    def __init__(self) -> None:
        self.locks: dict[str, Lock] = {}

    @abstractmethod
    def store_run_overview(self, overview: RunOverview) -> None:
        """Stores a :class:`RunOverview`.

        Args:
            overview: The overview to be persisted.
        """
        pass

    @abstractmethod
    def _create_temporary_run_data(self, tmp_hash: str, run_id: str) -> None:
        pass

    @abstractmethod
    def _delete_temporary_run_data(self, tmp_hash: str) -> None:
        pass

    @abstractmethod
    def _temp_store_finished_example(self, tmp_hash: str, example_id: str) -> None:
        pass

    @abstractmethod
    def finished_examples(self, tmp_hash: str) -> Optional[RecoveryData]:
        pass

    @final
    def create_temporary_run_data(self, tmp_hash: str, run_id: str) -> None:
        self.locks[tmp_hash] = lock()
        self._create_temporary_run_data(tmp_hash, run_id)

    @final
    def delete_temporary_run_data(self, tmp_hash: str) -> None:
        del self.locks[tmp_hash]
        self._delete_temporary_run_data(tmp_hash)

    @final
    def temp_store_finished_example(self, tmp_hash: str, example_id: str) -> None:
        with self.locks[tmp_hash]:
            self._temp_store_finished_example(tmp_hash, example_id)

    @abstractmethod
    def run_overview(self, run_id: str) -> Optional[RunOverview]:
        """Returns a :class:`RunOverview` for the given ID.

        Args:
            run_id: ID of the run overview to retrieve.

        Returns:
            :class:`RunOverview` if it was found, `None` otherwise.
        """
        ...

    def run_overviews(self) -> Iterable[RunOverview]:
        """Returns all :class:`RunOverview`s sorted by their ID.

        Yields:
            :class:`Iterable` of :class:`RunOverview`s.
        """
        for run_id in self.run_overview_ids():
            run_overview = self.run_overview(run_id)
            if run_overview is not None:
                yield run_overview

    @abstractmethod
    def run_overview_ids(self) -> Sequence[str]:
        """Returns sorted IDs of all stored :class:`RunOverview`s.

        Returns:
            A :class:`Sequence` of the :class:`RunOverview` IDs.
        """
        ...

    @abstractmethod
    def store_example_output(self, example_output: ExampleOutput[Output]) -> None:
        """Stores an :class:`ExampleOutput`.

        Args:
            example_output: The example output to be persisted.
        """
        ...

    @final
    def store_example_output_parallel(
        self, tmp_hash: str, example_output: ExampleOutput[Output]
    ) -> None:
        with self.locks[tmp_hash]:
            self.store_example_output(example_output)

    @abstractmethod
    def example_output(
        self, run_id: str, example_id: str, output_type: type[Output]
    ) -> Optional[ExampleOutput[Output] | ExampleOutput[FailedExampleRun]]:
        """Returns :class:`ExampleOutput` for the given run ID and example ID.

        Args:
            run_id: The ID of the linked run overview.
            example_id: ID of the example to retrieve.
            output_type: Type of output that the `Task` returned in :func:`Task.do_run`

        Returns:
            class:`ExampleOutput` if it was found, `None` otherwise.
        """
        ...

    @abstractmethod
    def example_outputs(
        self, run_id: str, output_type: type[Output]
    ) -> Iterable[ExampleOutput[Output] | ExampleOutput[FailedExampleRun]]:
        """Returns all :class:`ExampleOutput` for a given run ID sorted by their example ID.

        Args:
            run_id: The ID of the run overview.
            output_type: Type of output that the `Task` returned in :func:`Task.do_run`

        Returns:
            :class:`Iterable` of :class:`ExampleOutput`s.
        """
        ...

    @abstractmethod
    def example_output_ids(self, run_id: str) -> Sequence[str]:
        """Returns the sorted IDs of all :class:`ExampleOutput`s for a given run ID.

        Args:
            run_id: The ID of the run overview.

        Returns:
            A :class:`Sequence` of all :class:`ExampleOutput` IDs.
        """
        ...

    def successful_example_outputs(
        self, run_id: str, output_type: type[Output]
    ) -> Iterable[ExampleOutput[Output]]:
        """Returns all :class:`ExampleOutput` for successful example runs with a given run-overview ID sorted by their example ID.

        Args:
            run_id: The ID of the run overview.
            output_type: Type of output that the `Task` returned in :func:`Task.do_run`

        Returns:
            :class:`Iterable` of :class:`ExampleOutput`s.
        """
        results = self.example_outputs(run_id, output_type)
        return (r for r in results if not isinstance(r.output, FailedExampleRun))  # type: ignore

    def failed_example_outputs(
        self, run_id: str, output_type: type[Output]
    ) -> Iterable[ExampleOutput[FailedExampleRun]]:
        """Returns all :class:`ExampleOutput` for failed example runs with a given run-overview ID sorted by their example ID.

        Args:
            run_id: The ID of the run overview.
            output_type: Type of output that the `Task` returned in :func:`Task.do_run`

        Returns:
            :class:`Iterable` of :class:`ExampleOutput`s.
        """
        results = self.example_outputs(run_id, output_type)
        return (r for r in results if isinstance(r.output, FailedExampleRun))  # type: ignore

    @abstractmethod
    def example_tracer(self, run_id: str, example_id: str) -> Optional[Tracer]:
        """Returns an :class:`Optional[Tracer]` for the given run ID and example ID.

        Args:
            run_id: The ID of the linked run overview.
            example_id: ID of the example whose :class:`Tracer` should be retrieved.

        Returns:
            A :class:`Tracer` if it was found, `None` otherwise.
        """
        ...

    @abstractmethod
    def create_tracer_for_example(self, run_id: str, example_id: str) -> Tracer:
        """Creates and returns a :class:`Tracer` for the given run ID and example ID.

        Args:
            run_id: The ID of the linked run overview.
            example_id: ID of the example whose :class:`Tracer` should be retrieved.

        Returns:
            A :.class:`Tracer`.
        """
        ...



from collections.abc import Sequence
from pathlib import Path
from typing import Optional

from fsspec.implementations.local import LocalFileSystem  # type: ignore

from pharia_studio_sdk.evaluation.evaluation.domain import (
    Evaluation,
    EvaluationOverview,
    ExampleEvaluation,
    FailedExampleEvaluation,
    PartialEvaluationOverview,
)
from pharia_studio_sdk.evaluation.evaluation.evaluation_repository import (
    EvaluationRepository,
    SerializedExampleEvaluation,
)
from pharia_studio_sdk.evaluation.evaluation.evaluator.async_evaluator import (
    AsyncEvaluationRepository,
)
from pharia_studio_sdk.evaluation.infrastructure.file_system_based_repository import (
    FileSystemBasedRepository,
)


class FileSystemEvaluationRepository(EvaluationRepository, FileSystemBasedRepository):
    """An :class:`EvaluationRepository` that stores evaluation results in JSON files."""

    def store_evaluation_overview(self, overview: EvaluationOverview) -> None:
        self.write_utf8(
            self._evaluation_overview_path(overview.id),
            overview.model_dump_json(indent=2),
            create_parents=True,
        )
        # initialize the evaluation directory to make sure the evaluation "exists", even
        # if we did not store any examples in it
        self.mkdir(self._eval_directory(overview.id))

    def evaluation_overview(self, evaluation_id: str) -> Optional[EvaluationOverview]:
        file_path = self._evaluation_overview_path(evaluation_id)
        if not self.exists(file_path):
            return None

        content = self.read_utf8(file_path)
        return EvaluationOverview.model_validate_json(content)

    def evaluation_overview_ids(self) -> Sequence[str]:
        return sorted(self.file_names(self._eval_root_directory()))

    def store_example_evaluation(
        self, example_evaluation: ExampleEvaluation[Evaluation]
    ) -> None:
        serialized_result = SerializedExampleEvaluation.from_example_result(
            example_evaluation
        )
        self.write_utf8(
            self._example_result_path(
                example_evaluation.evaluation_id, example_evaluation.example_id
            ),
            serialized_result.model_dump_json(indent=2),
            create_parents=True,
        )

    def example_evaluation(
        self, evaluation_id: str, example_id: str, evaluation_type: type[Evaluation]
    ) -> Optional[
        ExampleEvaluation[Evaluation] | ExampleEvaluation[FailedExampleEvaluation]
    ]:
        file_path = self._example_result_path(evaluation_id, example_id)
        if not self.exists(file_path.parent):
            raise ValueError(
                f"Repository does not contain an evaluation with id: {evaluation_id}"
            )
        if not self.exists(file_path):
            return None

        content = self.read_utf8(file_path)
        serialized_example = SerializedExampleEvaluation.model_validate_json(content)
        return serialized_example.to_example_result(evaluation_type)

    def example_evaluations(
        self, evaluation_id: str, evaluation_type: type[Evaluation]
    ) -> Sequence[
        ExampleEvaluation[Evaluation] | ExampleEvaluation[FailedExampleEvaluation]
    ]:
        path = self._eval_directory(evaluation_id)
        if not self.exists(path):
            raise ValueError(
                f"Repository does not contain an evaluation with id: {evaluation_id}"
            )

        example_evaluations: list[
            ExampleEvaluation[Evaluation] | ExampleEvaluation[FailedExampleEvaluation]
        ] = []
        for file_name in self.file_names(path):
            evaluation = self.example_evaluation(
                evaluation_id, file_name, evaluation_type
            )
            if evaluation is not None:
                example_evaluations.append(evaluation)

        return sorted(example_evaluations, key=lambda i: i.example_id)

    def _eval_root_directory(self) -> Path:
        path = self._root_directory / "evaluations"
        return path

    def _eval_directory(self, evaluation_id: str) -> Path:
        path = self._eval_root_directory() / evaluation_id
        return path

    def _example_result_path(self, evaluation_id: str, example_id: str) -> Path:
        path = (self._eval_directory(evaluation_id) / example_id).with_suffix(".json")
        return path

    def _evaluation_overview_path(self, evaluation_id: str) -> Path:
        path = self._eval_directory(evaluation_id).with_suffix(".json")
        return path


class FileEvaluationRepository(FileSystemEvaluationRepository):
    def __init__(self, root_directory: Path) -> None:
        super().__init__(LocalFileSystem(), root_directory)

    @staticmethod
    def path_to_str(path: Path) -> str:
        return str(path)


class AsyncFileEvaluationRepository(
    FileEvaluationRepository, AsyncEvaluationRepository
):
    def store_partial_evaluation_overview(
        self, overview: PartialEvaluationOverview
    ) -> None:
        self.write_utf8(
            self._partial_evaluation_overview_path(overview.id),
            overview.model_dump_json(indent=2),
            create_parents=True,
        )
        # initialize the evaluation directory to make sure the evaluation "exists", even
        # if we did not store any examples in it
        self.mkdir(self._eval_directory(overview.id))

    def partial_evaluation_overview(
        self, evaluation_id: str
    ) -> Optional[PartialEvaluationOverview]:
        file_path = self._partial_evaluation_overview_path(evaluation_id)
        if not self.exists(file_path):
            return None

        content = self.read_utf8(file_path)
        return PartialEvaluationOverview.model_validate_json(content)

    def partial_evaluation_overview_ids(self) -> Sequence[str]:
        return sorted(self.file_names(self._eval_root_directory(), "partial_json"))

    def _partial_evaluation_overview_path(self, evaluation_id: str) -> Path:
        path = self._eval_directory(evaluation_id).with_suffix(".partial_json")
        return path



from abc import ABC
from collections.abc import Iterable, Mapping, Sequence
from functools import cached_property
from typing import (
    Generic,
    Optional,
    TypeVar,
    cast,
    get_args,
    get_origin,
)

from pharia_inference_sdk.core import Input, Output

from pharia_studio_sdk.evaluation.dataset.dataset_repository import DatasetRepository
from pharia_studio_sdk.evaluation.dataset.domain import Example, ExpectedOutput
from pharia_studio_sdk.evaluation.evaluation.domain import Evaluation
from pharia_studio_sdk.evaluation.evaluation.evaluation_repository import (
    EvaluationRepository,
)
from pharia_studio_sdk.evaluation.infrastructure.repository_navigator import (
    EvaluationLineage,
    RepositoryNavigator,
)
from pharia_studio_sdk.evaluation.run.domain import (
    ExampleOutput,
    FailedExampleRun,
    RunOverview,
    SuccessfulExampleOutput,
)
from pharia_studio_sdk.evaluation.run.run_repository import RunRepository


class EvaluationLogicBase(Generic[Input, Output, ExpectedOutput, Evaluation]):
    pass


class EvaluatorBase(Generic[Input, Output, ExpectedOutput, Evaluation], ABC):
    """Base class for Evaluators that can handle automatic evaluation scenarios.

    Provides methods for type inference and loading data from the repositories.

    Arguments:
        dataset_repository: The repository with the examples that will be taken for the evaluation.
        run_repository: The repository of the runs to evaluate.
        evaluation_repository: The repository that will be used to store evaluation results.
        description: Human-readable description for the evaluator.
        evaluation_logic: The logic to use for evaluation.

    Generics:
        Input: Type to be passed to the :class:`Task` as input. Part of `Example`.
        Output: Type of the output of the :class:`Task` to be evaluated. Part of `ExampleOutput`
        ExpectedOutput: Type that the `Output` will be compared against. Part of `Example`.
        Evaluation: Type of the metrics that come from the evaluated :class:`Task`. Part of `ExampleEvaluation`
    """

    def __init__(
        self,
        dataset_repository: DatasetRepository,
        run_repository: RunRepository,
        evaluation_repository: EvaluationRepository,
        description: str,
        evaluation_logic: EvaluationLogicBase[
            Input, Output, ExpectedOutput, Evaluation
        ],
    ) -> None:
        self._dataset_repository = dataset_repository
        self._run_repository = run_repository
        self._evaluation_repository = evaluation_repository
        self._evaluation_logic = evaluation_logic

        self.description = description

    @cached_property
    def _get_types(self) -> Mapping[str, type]:
        """Type magic function that gets the actual types of the generic parameters.

        Traverses the inheritance history of `EvaluationLogicBase`-subclasses to find an actual type every time a TypeVar is replaced.

        Returns:
            Name of generic parameter to the type found.
        """

        def is_eligible_subclass(parent: type) -> bool:
            return hasattr(parent, "__orig_bases__") and issubclass(
                parent, EvaluationLogicBase
            )

        def update_types() -> None:
            num_types_set = 0
            for current_index, current_type in enumerate(current_types):
                if type(current_type) is not TypeVar:
                    type_var_count = num_types_set - 1
                    final_element_index = -1
                    for element_index, element in enumerate(type_list):
                        final_element_index = element_index
                        if type(element) is TypeVar:
                            type_var_count += 1
                        if type_var_count == current_index:
                            break
                    assert type_var_count == current_index
                    type_list[final_element_index] = current_type
                    num_types_set += 1

        # mypy does not know __orig_bases__
        base_types = EvaluationLogicBase.__orig_bases__[0]  # type: ignore
        type_list: list[type | TypeVar] = list(get_args(base_types))
        possible_parent_classes = [
            p
            for p in reversed(type(self._evaluation_logic).__mro__)
            if is_eligible_subclass(p)
        ]
        for parent in possible_parent_classes:
            # mypy does not know __orig_bases__
            for base in parent.__orig_bases__:  # type: ignore
                origin = get_origin(base)
                if origin is None or not issubclass(origin, EvaluationLogicBase):
                    continue
                current_types = list(get_args(base))
                update_types()

        return {
            name: param_type
            for name, param_type in zip(
                (a.__name__ for a in get_args(base_types)), type_list, strict=True
            )
            if type(param_type) is not TypeVar
        }

    def input_type(self) -> type[Input]:
        """Returns the type of the evaluated task's input.

        This can be used to retrieve properly typed :class:`Example`s of a dataset
        from a :class:`DatasetRepository`.

        Returns:
            The type of the evaluated task's input.
        """
        try:
            input_type = self._get_types["Input"]
        except KeyError:
            raise TypeError(
                f"Alternatively overwrite input_type() in {type(self)}"
            ) from None
        return cast(type[Input], input_type)

    def output_type(self) -> type[Output]:
        """Returns the type of the evaluated task's output.

        This can be used to retrieve properly typed outputs of an evaluation run
        from a :class:`RunRepository`.

        Returns:
            The type of the evaluated task's output.
        """
        try:
            output_type = self._get_types["Output"]
        except KeyError:
            raise TypeError(
                f"Alternatively overwrite output_type() in {type(self)}"
            ) from None
        return cast(type[Output], output_type)

    def expected_output_type(self) -> type[ExpectedOutput]:
        """Returns the type of the evaluated task's expected output.

        This can be used to retrieve properly typed :class:`Example`s of a dataset
        from a :class:`DatasetRepository`.

        Returns:
            The type of the evaluated task's expected output.
        """
        try:
            expected_output_type = self._get_types["ExpectedOutput"]
        except KeyError:
            raise TypeError(
                f"Alternatively overwrite expected_output_type() in {type(self)}"
            ) from None
        return cast(type[ExpectedOutput], expected_output_type)

    def evaluation_type(self) -> type[Evaluation]:
        """Returns the type of the evaluation result of an example.

        This can be used to retrieve properly typed evaluations of an evaluation run
        from an :class:`EvaluationRepository`

        Returns:
            Returns the type of the evaluation result of an example.
        """
        try:
            evaluation_type = self._get_types["Evaluation"]
        except KeyError:
            raise TypeError(
                f"Alternatively overwrite evaluation_type() in {type(self)}"
            ) from None
        return cast(type[Evaluation], evaluation_type)

    def _load_run_overviews(self, *run_ids: str) -> set[RunOverview]:
        if not run_ids:
            raise ValueError("At least one run-id needs to be provided")
        run_overviews = set()
        for run_id in run_ids:
            run_overview = self._run_repository.run_overview(run_id)
            if not run_overview:
                raise ValueError(f"No RunOverview found for run-id: {run_id}")

            run_overviews.add(run_overview)
        return run_overviews

    def _raise_if_overviews_have_different_dataset(
        self, run_overviews: set[RunOverview]
    ) -> None:
        if any(
            next(iter(run_overviews)).dataset_id != run_overview.dataset_id
            for run_overview in run_overviews
        ):
            raise ValueError(
                f"All run overviews must reference the same dataset: {run_overviews}"
            )

    def _raise_if_overviews_have_different_number_of_runs(
        self, run_overviews: set[RunOverview]
    ) -> None:
        run_overview_list = list(run_overviews)
        if any(
            run_overview_list[0].failed_example_count
            + run_overview_list[0].successful_example_count
            != run_overview.successful_example_count + run_overview.failed_example_count
            for run_overview in run_overview_list
        ):
            raise ValueError(
                f"All run overviews must contain the same number of items: {run_overviews}"
            )

    def _retrieve_example_outputs(
        self, run_overviews: set[RunOverview]
    ) -> Iterable[tuple[ExampleOutput[Output], ...]]:
        # this uses the assumption that the example outputs are sorted and there is never a missing example
        example_outputs_for_example: Iterable[tuple[ExampleOutput[Output], ...]] = zip(
            *(
                self._run_repository.example_outputs(
                    run_overview.id, self.output_type()
                )
                for run_overview in run_overviews
            ),
            strict=True,
        )

        return example_outputs_for_example

    def _retrieve_examples(
        self, dataset_id: str
    ) -> Iterable[Example[Input, ExpectedOutput]]:
        examples = self._dataset_repository.examples(
            dataset_id,
            self.input_type(),
            self.expected_output_type(),
        )
        if examples is None:
            raise ValueError(f"Dataset: {dataset_id} not found")

        return examples

    def _generate_evaluation_inputs(
        self,
        examples: Iterable[Example[Input, ExpectedOutput]],
        example_outputs_for_example: Iterable[tuple[ExampleOutput[Output], ...]],
        skip_example_on_any_failure: bool,
        num_examples: Optional[int],
    ) -> Iterable[
        tuple[
            Example[Input, ExpectedOutput],
            Sequence[SuccessfulExampleOutput[Output]],
        ]
    ]:
        current_example = 0

        for example, example_outputs in zip(
            examples, example_outputs_for_example, strict=False
        ):
            if any(
                example.id != example_output.example_id
                for example_output in example_outputs
            ):
                raise ValueError(
                    "The ids of example and output do not match. Therefore, the evaluation cannot continue.\n"
                    + f"example id: {example.id}, output id: {example_outputs}."
                )
            if skip_example_on_any_failure and any(
                isinstance(output.output, FailedExampleRun)
                for output in example_outputs
            ):
                continue

            successful_example_outputs = [
                cast(SuccessfulExampleOutput[Output], output)
                for output in example_outputs
                if not isinstance(output.output, FailedExampleRun)
            ]

            if num_examples and current_example >= num_examples:
                break
            current_example += 1

            yield (
                example,
                successful_example_outputs,
            )

    def _retrieve_eval_logic_input(
        self,
        run_overviews: set[RunOverview],
        skip_example_on_any_failure: bool,
        num_examples: Optional[int] = None,
    ) -> Iterable[
        tuple[
            Example[Input, ExpectedOutput],
            Sequence[SuccessfulExampleOutput[Output]],
        ]
    ]:
        """Create pairings of :class:`Example` and all corresponding :class:`ExampleOutputs`.

        In case an Example is matched with a FailedExampleRun, that example is skipped, even if
        there are other successful ExampleOutputs present for this example.

        Args:
            run_overviews: Run overviews to gather data from.
            skip_example_on_any_failure: Skip example on any failure.
            num_examples: Maximum amount of examples to gather. Defaults to None.

        Returns:
            Iterable over pairs of :class:`Example` and all corresponding :class:`ExampleOutputs`.
        """
        self._raise_if_overviews_have_different_dataset(run_overviews)
        self._raise_if_overviews_have_different_number_of_runs(run_overviews)
        example_outputs_for_example = self._retrieve_example_outputs(run_overviews)
        dataset_id = next(iter(run_overviews)).dataset_id
        examples = self._retrieve_examples(dataset_id)
        return self._generate_evaluation_inputs(
            examples,
            example_outputs_for_example,
            skip_example_on_any_failure,
            num_examples,
        )

    def failed_evaluations(
        self, evaluation_id: str
    ) -> Iterable[EvaluationLineage[Input, ExpectedOutput, Output, Evaluation]]:
        """Returns the `EvaluationLineage` objects for all failed example evaluations that belong to the given evaluation ID.

        Args:
            evaluation_id: The ID of the evaluation overview

        Returns:
            :class:`Iterable` of :class:`EvaluationLineage`s.
        """
        failed_example_evaluations = (
            self._evaluation_repository.failed_example_evaluations(
                evaluation_id, evaluation_type=self.evaluation_type()
            )
        )
        lineages = (
            self.evaluation_lineage(evaluation_id, output.example_id)
            for output in failed_example_evaluations
        )
        return (lineage for lineage in lineages if lineage is not None)

    def evaluation_lineages(
        self, evaluation_id: str
    ) -> Iterable[EvaluationLineage[Input, ExpectedOutput, Output, Evaluation]]:
        """Wrapper for `RepositoryNagivator.evaluation_lineages`.

        Args:
            evaluation_id: The id of the evaluation

        Returns:
            An iterator over all :class:`EvaluationLineage`s for the given evaluation id.
        """
        navigator = RepositoryNavigator(
            self._dataset_repository, self._run_repository, self._evaluation_repository
        )
        return navigator.evaluation_lineages(
            evaluation_id=evaluation_id,
            input_type=self.input_type(),
            expected_output_type=self.expected_output_type(),
            output_type=self.output_type(),
            evaluation_type=self.evaluation_type(),
        )

    def evaluation_lineage(
        self, evaluation_id: str, example_id: str
    ) -> EvaluationLineage[Input, ExpectedOutput, Output, Evaluation] | None:
        """Wrapper for `RepositoryNagivator.evaluation_lineage`.

        Args:
            evaluation_id: The id of the evaluation
            example_id: The id of the example of interest

        Returns:
            The :class:`EvaluationLineage` for the given evaluation id and example id.
            Returns `None` if the lineage is not complete because either an example, a run, or an evaluation does not exist.
        """
        navigator = RepositoryNavigator(
            self._dataset_repository, self._run_repository, self._evaluation_repository
        )
        return navigator.evaluation_lineage(
            evaluation_id=evaluation_id,
            example_id=example_id,
            input_type=self.input_type(),
            expected_output_type=self.expected_output_type(),
            output_type=self.output_type(),
            evaluation_type=self.evaluation_type(),
        )



from collections.abc import Sequence
from typing import Any, Optional

from pharia_studio_sdk.evaluation.aggregation.aggregation_repository import (
    AggregationRepository,
)
from pharia_studio_sdk.evaluation.aggregation.domain import (
    AggregatedEvaluation,
    AggregationOverview,
)


class InMemoryAggregationRepository(AggregationRepository):
    def __init__(self) -> None:
        super().__init__()
        self._aggregation_overviews: dict[str, AggregationOverview[Any]] = dict()

    def store_aggregation_overview(
        self, aggregation_overview: AggregationOverview[AggregatedEvaluation]
    ) -> None:
        self._aggregation_overviews[aggregation_overview.id] = aggregation_overview

    def aggregation_overview(
        self, aggregation_id: str, aggregation_type: type[AggregatedEvaluation]
    ) -> Optional[AggregationOverview[AggregatedEvaluation]]:
        overview = self._aggregation_overviews.get(aggregation_id, None)
        if overview is None or type(overview.statistics) is aggregation_type:
            return overview
        return AggregationOverview[AggregatedEvaluation](
            evaluation_overviews=overview.evaluation_overviews,
            id=overview.id,
            start=overview.start,
            end=overview.end,
            successful_evaluation_count=overview.successful_evaluation_count,
            crashed_during_evaluation_count=overview.crashed_during_evaluation_count,
            description=overview.description,
            statistics=aggregation_type.model_validate(overview.statistics),
            labels=overview.labels,
            metadata=overview.metadata,
        )

    def aggregation_overview_ids(self) -> Sequence[str]:
        return sorted(list(self._aggregation_overviews.keys()))



from collections.abc import Sequence
from pathlib import Path
from typing import Optional

from fsspec.implementations.local import LocalFileSystem  # type: ignore

from pharia_studio_sdk.evaluation.aggregation.aggregation_repository import (
    AggregationRepository,
)
from pharia_studio_sdk.evaluation.aggregation.domain import (
    AggregatedEvaluation,
    AggregationOverview,
)
from pharia_studio_sdk.evaluation.infrastructure.file_system_based_repository import (
    FileSystemBasedRepository,
)


class FileSystemAggregationRepository(AggregationRepository, FileSystemBasedRepository):
    _SUB_DIRECTORY = "aggregations"

    def store_aggregation_overview(
        self, aggregation_overview: AggregationOverview[AggregatedEvaluation]
    ) -> None:
        self.write_utf8(
            self._aggregation_overview_path(aggregation_overview.id),
            aggregation_overview.model_dump_json(indent=2),
            create_parents=True,
        )

    def aggregation_overview(
        self, aggregation_id: str, aggregation_type: type[AggregatedEvaluation]
    ) -> Optional[AggregationOverview[AggregatedEvaluation]]:
        file_path = self._aggregation_overview_path(aggregation_id)

        if not self.exists(file_path):
            return None

        content = self.read_utf8(file_path)
        return AggregationOverview[aggregation_type].model_validate_json(  # type:ignore
            content
        )

    def aggregation_overview_ids(self) -> Sequence[str]:
        return sorted(self.file_names(self._aggregation_root_directory()))

    def _aggregation_root_directory(self) -> Path:
        return self._root_directory / self._SUB_DIRECTORY

    def _aggregation_directory(self, evaluation_id: str) -> Path:
        return self._aggregation_root_directory() / evaluation_id

    def _aggregation_overview_path(self, aggregation_id: str) -> Path:
        return self._aggregation_directory(aggregation_id).with_suffix(".json")


class FileAggregationRepository(FileSystemAggregationRepository):
    def __init__(self, root_directory: Path) -> None:
        super().__init__(LocalFileSystem(), root_directory)

    @staticmethod
    def path_to_str(path: Path) -> str:
        return str(path)



import gzip
import json
import os
from collections import defaultdict, deque
from collections.abc import Iterable, Sequence
from datetime import datetime
from typing import Any, Generic, Optional, TypeVar
from urllib.parse import urljoin
from uuid import UUID, uuid4

import requests
from pharia_inference_sdk.core.tracer.tracer import (
    ExportedSpan,
    ExportedSpanList,
    PydanticSerializable,
    Tracer,
)
from pydantic import BaseModel, Field, RootModel, field_validator
from requests.exceptions import ConnectionError, MissingSchema

from pharia_studio_sdk.connectors.base.json_serializable import (
    JsonSerializable,
    SerializableDict,
)

Input = TypeVar("Input", bound=PydanticSerializable)
ExpectedOutput = TypeVar("ExpectedOutput", bound=PydanticSerializable)
Output = TypeVar("Output", bound=PydanticSerializable)
Evaluation = TypeVar("Evaluation", bound=BaseModel, covariant=True)


class StudioProject(BaseModel):
    name: str
    description: Optional[str]


class StudioExample(BaseModel, Generic[Input, ExpectedOutput]):
    """Represents an instance of :class:`Example`as sent to Studio.

    Attributes:
        input: Input for the :class:`Task`. Has to be same type as the input for the task used.
        expected_output: The expected output from a given example run.
            This will be used by the evaluator to compare the received output with.
        id: Identifier for the example, defaults to uuid.
        metadata: Optional dictionary of custom key-value pairs.

    Generics:
        Input: Interface to be passed to the :class:`Task` that shall be evaluated.
        ExpectedOutput: Output that is expected from the run with the supplied input.
    """

    input: Input
    expected_output: ExpectedOutput
    id: str = Field(default_factory=lambda: str(uuid4()))
    metadata: Optional[SerializableDict] = None


class StudioDataset(BaseModel):
    """Represents a :class:`Dataset` linked to multiple examples as sent to Studio.

    Attributes:
        id: Dataset ID.
        name: A short name of the dataset.
        label: Labels for filtering datasets. Defaults to empty list.
        metadata: Additional information about the dataset. Defaults to empty dict.
    """

    id: str = Field(default_factory=lambda: str(uuid4()))
    name: str
    labels: set[str] = set()
    metadata: SerializableDict = dict()


class EvaluationLogicIdentifier(BaseModel):
    logic: str  # code from the evaluation logic as a string
    input_schema: dict[str, Any]
    output_schema: dict[str, Any]
    expected_output_schema: dict[str, Any]
    evaluation_schema: dict[str, Any]


class AggregationLogicIdentifier(BaseModel):
    logic: str  # code from the aggregation logic as a string
    evaluation_schema: dict[str, Any]
    aggregation_schema: dict[str, Any]


class PostBenchmarkRequest(BaseModel):
    dataset_id: str
    name: str
    description: Optional[str]
    benchmark_metadata: Optional[dict[str, Any]]
    evaluation_logic: EvaluationLogicIdentifier
    aggregation_logic: AggregationLogicIdentifier


class GetBenchmarkResponse(BaseModel):
    id: str
    project_id: str
    dataset_id: str
    name: str
    description: str | None
    benchmark_metadata: dict[str, Any] | None
    evaluation_logic: EvaluationLogicIdentifier
    aggregation_logic: AggregationLogicIdentifier
    created_at: datetime
    updated_at: datetime | None
    last_executed_at: datetime | None
    created_by: str | None
    updated_by: str | None

    @field_validator("project_id", mode="before")
    def transform_id_to_str(cls, value) -> str:
        if type(value) is int or type(value) is UUID:
            return str(value)
        return value


class PostBenchmarkExecution(BaseModel):
    name: str
    description: Optional[str]
    labels: Optional[set[str]]
    metadata: Optional[dict[str, Any]]
    start: datetime
    end: datetime
    # Run Overview
    run_start: datetime
    run_end: datetime
    run_successful_count: int
    run_failed_count: int
    run_success_avg_latency: float
    run_success_avg_token_count: float
    # Eval Overview
    eval_start: datetime
    eval_end: datetime
    eval_successful_count: int
    eval_failed_count: int
    # Aggregation Overview
    aggregation_start: datetime
    aggregation_end: datetime
    statistics: JsonSerializable


class GetDatasetExamplesResponse(BaseModel, Generic[Input, ExpectedOutput]):
    total: int
    page: int
    size: int
    num_pages: int
    items: Sequence[StudioExample[Input, ExpectedOutput]]


class BenchmarkLineage(BaseModel, Generic[Input, ExpectedOutput, Output, Evaluation]):
    trace_id: str
    input: Input
    expected_output: ExpectedOutput
    output: Output
    example_metadata: Optional[dict[str, Any]] = None
    evaluation: Any
    run_latency: int
    run_tokens: int


class PostBenchmarkLineagesRequest(RootModel[Sequence[BenchmarkLineage]]):
    pass


class PostBenchmarkLineagesResponse(RootModel[Sequence[str]]):
    pass


class GetBenchmarkLineageResponse(BaseModel):
    id: str
    trace_id: str
    benchmark_execution_id: str
    input: JsonSerializable
    expected_output: JsonSerializable
    example_metadata: Optional[dict[str, JsonSerializable]] = None
    output: JsonSerializable
    evaluation: JsonSerializable
    run_latency: int
    run_tokens: int


class StudioClient:
    """Client for communicating with Studio.

    Attributes:
      project_id: The unique identifier of the project currently in use.
      url: The url of your current Studio instance.
    """

    @staticmethod
    def get_headers(auth_token: Optional[str] = None) -> dict[str, str]:
        _token = auth_token if auth_token is not None else os.getenv("AA_TOKEN")
        if _token is None:
            raise ValueError(
                "'AA_TOKEN' is not set and auth_token is not given as a parameter. Please provide one or the other."
            )
        return {
            "Accept": "application/json",
            "Authorization": f"Bearer {_token}",
        }

    @staticmethod
    def get_url(studio_url: Optional[str] = None) -> str:
        temp_url = studio_url if studio_url is not None else os.getenv("STUDIO_URL")
        if temp_url is None:
            raise ValueError(
                "'STUDIO_URL' is not set and url is not given as a parameter. Please provide one or the other."
            )
        return temp_url

    def __init__(
        self,
        project: str,
        studio_url: Optional[str] = None,
        auth_token: Optional[str] = None,
        create_project: bool = False,
    ) -> None:
        """Initializes the client.

        Runs a health check to check for a valid url of the Studio connection.
        It does not check for a valid authentication token, which happens later.

        Args:
            project: The human readable identifier provided by the user.
            studio_url: The url of your current Studio instance.
            auth_token: The authorization bearer token of the user. This corresponds to the user's Aleph Alpha token.
            create_project: If True, the client will try to create the project if it does not exist. Defaults to False.
        """
        self._headers = StudioClient.get_headers(auth_token)
        self.url = StudioClient.get_url(studio_url)
        self._check_connection()
        self._project_name = project
        self._project_id: str | None = self._get_project(self._project_name)

        if create_project and self._project_id is None:
            self._project_id = self.create_project(
                self._project_name, reuse_existing=True
            )

    def _check_connection(self) -> None:
        try:
            url = urljoin(self.url, "/health")
            response = requests.get(
                url,
                headers=self._headers,
            )
            response.raise_for_status()
        except MissingSchema:
            raise ValueError(
                "The given url of the studio client is invalid. Make sure to include http:// in your url."
            ) from None
        except ConnectionError:
            raise ValueError(
                "The given url of the studio client does not point to a server."
            ) from None
        except requests.HTTPError:
            raise ValueError(
                f"The given url of the studio client does not point to a healthy studio: {response.status_code}: {response.json()}"
            ) from None

    @property
    def project_id(self) -> str:
        if self._project_id is None:
            project_id = self._get_project(self._project_name)
            if project_id is None:
                raise ValueError(
                    f"Project {self._project_name} was not available. Consider creating it with `StudioClient.create_project`."
                )
            self._project_id = project_id
        return self._project_id

    def _get_project(self, project_name: str) -> str | None:
        url = urljoin(self.url, "/api/projects")
        response = requests.get(
            url,
            headers=self._headers,
        )
        response.raise_for_status()
        all_projects = response.json()

        matching_projects = [
            proj for proj in all_projects if proj["name"] == project_name
        ]

        if not matching_projects:
            return None
        if len(matching_projects) > 1:
            raise ValueError(
                f"Multiple projects with name '{project_name}' found. Please make sure project names available to you are unique. E.g. rename/delete the duplicated project"
            )

        project_of_interest = matching_projects[0]
        # Studio API service < v0.1.0 does not have a "project_id" field
        if "project_id" in project_of_interest:
            return str(project_of_interest["project_id"])
        return str(project_of_interest["id"])

    def create_project(
        self,
        project: str,
        description: Optional[str] = None,
        reuse_existing: bool = False,
    ) -> str:
        """Creates a project in Studio.

        Projects are uniquely identified by the user provided name.

        Args:
            project: User provided name of the project.
            description: Description explaining the usage of the project. Defaults to None.
            reuse_existing: Reuse project with specified name if already existing. Defaults to False.


        Returns:
            The ID of the newly created project.
        """
        if reuse_existing:
            fetched_project = self._get_project(project)
            if fetched_project is not None:
                return fetched_project
        url = urljoin(self.url, "/api/projects")
        data = StudioProject(name=project, description=description)
        response = requests.post(
            url,
            data=data.model_dump_json(),
            headers=self._headers,
        )
        response.raise_for_status()
        return str(response.json())

    def _delete_project(
        self,
        project_id: str,
    ) -> None:
        url = urljoin(self.url, f"/api/projects/{project_id}")
        response = requests.delete(
            url,
            headers=self._headers,
        )
        response.raise_for_status()

    def submit_trace(self, data: Sequence[ExportedSpan]) -> str:
        """Sends the provided spans to Studio as a singular trace.

        The method fails if the span list is empty, has already been created or if
        spans belong to multiple traces.

        Args:
            data: :class:`Spans` to create the trace from. Created by exporting from a :class:`Tracer`.

        Returns:
            The ID of the created trace.
        """
        if len(data) == 0:
            raise ValueError("Tried to upload an empty trace")
        return self._upload_trace(ExportedSpanList(data))

    def submit_from_tracer(self, tracer: Tracer) -> list[str]:
        """Sends all trace data from the Tracer to Studio.

        Args:
            tracer: :class:`Tracer` to extract data from.

        Returns:
            List of created trace IDs.
        """
        traces = defaultdict(list)
        for span in tracer.export_for_viewing():
            traces[span.context.trace_id].append(span)

        return [self.submit_trace(value) for value in traces.values()]

    def _upload_trace(self, trace: ExportedSpanList) -> str:
        url = urljoin(self.url, f"/api/projects/{self.project_id}/traces")
        response = requests.post(
            url,
            data=trace.model_dump_json(),
            headers=self._headers,
        )
        match response.status_code:
            case 409:
                raise ValueError(
                    f"Trace with id {trace.root[0].context.trace_id} already exists."
                )
            case 422:
                raise ValueError(
                    f"Uploading the trace failed with 422. Response: {response.json()}"
                )
            case _:
                response.raise_for_status()
        return str(response.json())

    def submit_dataset(
        self,
        dataset: StudioDataset,
        examples: Iterable[StudioExample[Input, ExpectedOutput]],
    ) -> str:
        """Submits a dataset to Studio.

        Args:
            dataset: :class:`Dataset` to be uploaded
            examples: :class:`Examples` of the :class:`Dataset`

        Returns:
            ID of the created dataset
        """
        url = urljoin(self.url, f"/api/projects/{self.project_id}/evaluation/datasets")
        source_data_list = [example.model_dump_json() for example in examples]

        source_data_file = "\n".join(source_data_list).encode()

        data = {
            "name": dataset.name,
            "labels": list(dataset.labels) if dataset.labels is not None else [],
            "total_datapoints": len(source_data_list),
        }

        if dataset.metadata:
            data["metadata"] = json.dumps(dataset.metadata)

        response = requests.post(
            url,
            files={"source_data": source_data_file},
            data=data,
            headers=self._headers,
        )

        self._raise_for_status(response)
        return str(response.json())

    def get_dataset_examples(
        self,
        dataset_id: str,
        input_type: type[Input],
        expected_output_type: type[ExpectedOutput],
    ) -> Iterable[StudioExample[Input, ExpectedOutput]]:
        buffer_size = 200
        page_size = 100
        page: int | None = 1
        buffer: deque[StudioExample[Input, ExpectedOutput]] = deque()

        while True:
            if len(buffer) < buffer_size // 2 and page is not None:
                page_url = urljoin(
                    self.url,
                    f"/api/projects/{self.project_id}/evaluation/datasets/{dataset_id}/datapoints?page={page}&size={page_size}",
                )

                response = requests.get(page_url, headers=self._headers)

                if response.status_code == 200:
                    examples = GetDatasetExamplesResponse(**response.json()).items
                    buffer.extend(examples)

                    if len(examples) < page_size:
                        page = None
                    else:
                        page += 1
                else:
                    raise Exception(
                        f"Failed to fetch items from {page_url}. Status code: {response.status_code}"
                    )

            if len(buffer) > 0:
                yield StudioExample[
                    input_type, expected_output_type  # type: ignore
                ].model_validate_json(json_data=buffer.popleft().model_dump_json())
            else:
                if page is None:
                    break

    def submit_benchmark(
        self,
        dataset_id: str,
        eval_logic: EvaluationLogicIdentifier,
        aggregation_logic: AggregationLogicIdentifier,
        name: str,
        description: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
    ) -> str:
        benchmark = PostBenchmarkRequest(
            dataset_id=dataset_id,
            name=name,
            description=description,
            benchmark_metadata=metadata,
            evaluation_logic=eval_logic,
            aggregation_logic=aggregation_logic,
        )
        url = urljoin(
            self.url, f"/api/projects/{self.project_id}/evaluation/benchmarks"
        )
        response = requests.post(
            url,
            data=benchmark.model_dump_json(),
            headers=self._headers,
        )
        self._raise_for_status(response)
        return str(response.json())

    def get_benchmark(
        self,
        benchmark_id: str,
    ) -> GetBenchmarkResponse | None:
        url = urljoin(
            self.url,
            f"/api/projects/{self.project_id}/evaluation/benchmarks/{benchmark_id}",
        )
        response = requests.get(
            url,
            headers=self._headers,
        )
        self._raise_for_status(response)
        response_text = response.json()
        if response_text is None:
            return None
        return GetBenchmarkResponse.model_validate(response_text)

    def submit_benchmark_execution(
        self, benchmark_id: str, data: PostBenchmarkExecution
    ) -> str:
        url = urljoin(
            self.url,
            f"/api/projects/{self.project_id}/evaluation/benchmarks/{benchmark_id}/executions",
        )

        response = requests.post(
            url, headers=self._headers, data=data.model_dump_json()
        )

        self._raise_for_status(response)
        return str(response.json())

    def submit_benchmark_lineages(
        self,
        benchmark_lineages: Sequence[BenchmarkLineage],
        benchmark_id: str,
        execution_id: str,
        max_payload_size: int = 50
        * 1024
        * 1024,  # Maximum request size handled by Studio
    ) -> PostBenchmarkLineagesResponse:
        """Submit benchmark lineages in batches to avoid exceeding the maximum payload size.

        Args:
            benchmark_lineages: List of :class: `BenchmarkLineages` to submit.
            benchmark_id: ID of the benchmark.
            execution_id: ID of the execution.
            max_payload_size: Maximum size of the payload in bytes. Defaults to 50MB.

        Returns:
            Response containing the results of the submissions.
        """
        all_responses = []
        remaining_lineages = list(benchmark_lineages)
        lineage_sizes = [
            len(lineage.model_dump_json().encode("utf-8"))
            for lineage in benchmark_lineages
        ]

        while remaining_lineages:
            batch = []
            current_size = 0
            # Build batch while checking size
            for lineage, size in zip(remaining_lineages, lineage_sizes, strict=True):
                if current_size + size <= max_payload_size:
                    batch.append(lineage)
                    current_size += size
                else:
                    break

            if batch:
                # Send batch
                response = self._send_compressed_batch(
                    batch, benchmark_id, execution_id
                )
                all_responses.extend(response)

            else:  # Only reached if a lineage is too big for the request
                print("Lineage exceeds maximum of upload size", lineage)
                batch.append(lineage)
            remaining_lineages = remaining_lineages[len(batch) :]
            lineage_sizes = lineage_sizes[len(batch) :]

        return PostBenchmarkLineagesResponse(all_responses)

    def get_benchmark_lineage(
        self, benchmark_id: str, execution_id: str, lineage_id: str
    ) -> GetBenchmarkLineageResponse | None:
        url = urljoin(
            self.url,
            f"/api/projects/{self.project_id}/evaluation/benchmarks/{benchmark_id}/executions/{execution_id}/lineages/{lineage_id}",
        )
        response = requests.get(
            url,
            headers=self._headers,
        )
        self._raise_for_status(response)
        response_text = response.json()
        if response_text is None:
            return None
        return GetBenchmarkLineageResponse.model_validate(response_text)

    def _send_compressed_batch(
        self, batch: list[BenchmarkLineage], benchmark_id: str, execution_id: str
    ) -> list[str]:
        url = urljoin(
            self.url,
            f"/api/projects/{self.project_id}/evaluation/benchmarks/{benchmark_id}/executions/{execution_id}/lineages",
        )

        json_data = PostBenchmarkLineagesRequest(root=batch).model_dump_json()
        compressed_data = gzip.compress(json_data.encode("utf-8"))

        headers = {**self._headers, "Content-Encoding": "gzip"}

        response = requests.post(
            url,
            headers=headers,
            data=compressed_data,
        )

        self._raise_for_status(response)
        return response.json()

    def _raise_for_status(self, response: requests.Response) -> None:
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            print(
                f"The following error has been raised via execution {e.response.text}"
            )
            raise e



from argparse import ArgumentParser, Namespace
from collections.abc import Sequence
from importlib import import_module
from pathlib import Path
from sys import argv
from typing import Any

from dotenv import load_dotenv
from pharia_inference_sdk.connectors.limited_concurrency_client import (
    LimitedConcurrencyClient,
)

from pharia_studio_sdk.evaluation.aggregation.aggregator import Aggregator
from pharia_studio_sdk.evaluation.aggregation.file_aggregation_repository import (
    FileAggregationRepository,
)
from pharia_studio_sdk.evaluation.dataset.file_dataset_repository import (
    FileDatasetRepository,
)
from pharia_studio_sdk.evaluation.evaluation.evaluator.evaluator import Evaluator
from pharia_studio_sdk.evaluation.evaluation.file_evaluation_repository import (
    FileEvaluationRepository,
)
from pharia_studio_sdk.evaluation.run.file_run_repository import FileRunRepository
from pharia_studio_sdk.evaluation.run.runner import Runner


def function_from_string(fully_qualified_function_name: str) -> Any:
    mod_name, func_name = fully_qualified_function_name.rsplit(".", 1)
    mod = import_module(mod_name)
    return getattr(mod, func_name)


def create_task(factory: Any) -> Any:
    try:
        return factory()
    except TypeError:
        return factory(LimitedConcurrencyClient.from_env())


def parse_args(cli_args: Sequence[str]) -> Namespace:
    parser = ArgumentParser(description="Runs the given evaluation")
    parser.add_argument(
        "--eval-logic",
        required=True,
        type=function_from_string,
        help="A factory function for the evaluation logic. ",
    )
    parser.add_argument(
        "--aggregation-logic",
        required=True,
        type=function_from_string,
        help="A factory function for the aggregation logic. ",
    )
    parser.add_argument(
        "--task",
        required=True,
        type=function_from_string,
        help="A factory function for the task to be evaluated. "
        "This function can either take no parameter or an Aleph Alpha client as only parameter. "
        "If this corresponds to the init-parameters of the Evaluator "
        "the class-type can actually be provided as argument.",
    )
    parser.add_argument(
        "--dataset-repository-path",
        required=True,
        type=Path,
        help="Path to a file dataset repository.",
    )
    parser.add_argument(
        "--dataset-id",
        required=True,
        type=str,
        help="ID of a dataset that exists in the file dataset repository provided.",
    )
    parser.add_argument(
        "--target-dir",
        required=True,
        type=Path,
        help="Path to a directory where the evaluation results are stored. "
        "The directory is created if it does not exist. "
        "The process must have corresponding write permissions.",
    )
    parser.add_argument(
        "--description",
        required=True,
        type=str,
        help="Description of the evaluator.",
    )
    args = parser.parse_args(cli_args[1:])
    return args


def main(cli_args: Sequence[str]) -> None:
    args = parse_args(cli_args)
    dataset_repository = FileDatasetRepository(args.dataset_repository_path)
    runner_repository = FileRunRepository(args.target_dir)
    evaluation_repository = FileEvaluationRepository(args.target_dir)
    aggregation_repository = FileAggregationRepository(args.target_dir)
    description = args.description
    task = create_task(args.task)
    runner = Runner(task, dataset_repository, runner_repository, args.task.__name__)
    dataset_id = args.dataset_id
    run_overview_id = runner.run_dataset(dataset_id).id
    eval_logic = args.eval_logic()
    aggregation_logic = args.aggregation_logic()

    evaluator = Evaluator(
        dataset_repository,
        runner_repository,
        evaluation_repository,
        description,
        eval_logic,
    )
    aggregator = Aggregator(
        evaluation_repository, aggregation_repository, description, aggregation_logic
    )
    evaluation_overview = evaluator.evaluate_runs(run_overview_id)
    aggregator.aggregate_evaluation(evaluation_overview.id)


if __name__ == "__main__":
    load_dotenv()
    main(argv)



from pharia_studio_sdk.evaluation.aggregation.file_aggregation_repository import (
    FileSystemAggregationRepository,
)
from pharia_studio_sdk.evaluation.infrastructure.hugging_face_repository import (
    HuggingFaceRepository,
)


class HuggingFaceAggregationRepository(
    FileSystemAggregationRepository, HuggingFaceRepository
):
    # this class inherits all its behavior from its parents
    pass



from collections.abc import Sequence
from pathlib import Path
from typing import cast

from fsspec import AbstractFileSystem  # type: ignore


class FileSystemBasedRepository:
    """An :class:`FileBasedRepository` that stores evaluation results in files.

    Args:
        file_system: The specific file system to use from fsspec.
        root_directory: The folder where the files are stored. The folder
            (along with its parents) will be created if it does not exist yet.
    """

    def __init__(self, file_system: AbstractFileSystem, root_directory: Path) -> None:
        self._root_directory = root_directory
        self._file_system = file_system
        self.mkdir(root_directory)

    def write_utf8(
        self, path: Path, content: str, create_parents: bool = False
    ) -> None:
        if create_parents:
            self.mkdir(path.parent)
        self._file_system.write_text(self.path_to_str(path), content, encoding="utf-8")

    def read_utf8(self, path: Path) -> str:
        return cast(
            str, self._file_system.read_text(self.path_to_str(path), encoding="utf-8")
        )

    def remove_file(self, path: Path) -> None:
        self._file_system.rm_file(path)

    def exists(self, path: Path) -> bool:
        return cast(bool, self._file_system.exists(self.path_to_str(path)))

    def mkdir(self, path: Path) -> None:
        if self.exists(path):
            return
        try:
            self._file_system.makedir(self.path_to_str(path), create_parents=True)
        except FileExistsError:
            return

    def file_names(self, path: Path, file_type: str = "json") -> Sequence[str]:
        files = [
            Path(file)
            for file in self._file_system.ls(self.path_to_str(path), detail=False)
        ]
        return [file.stem for file in files if file.suffix == "." + file_type]

    @staticmethod
    def path_to_str(path: Path) -> str:
        """Returns a string for the given Path so that it's readable for the respective file system.

        Args:
            path: Given Path that should be converted.

        Returns:
            String representation of the given Path.
        """
        return str(path)



from collections.abc import Iterable, Sequence
from typing import Any, Optional, TypeVar, cast

from pharia_inference_sdk.core import Input, PydanticSerializable
from pydantic import TypeAdapter

from pharia_studio_sdk.connectors.base.json_serializable import (
    SerializableDict,
)
from pharia_studio_sdk.evaluation.dataset.dataset_repository import DatasetRepository
from pharia_studio_sdk.evaluation.dataset.domain import (
    Dataset,
    Example,
    ExpectedOutput,
)


class InMemoryDatasetRepository(DatasetRepository):
    def __init__(self) -> None:
        self._datasets_and_examples: dict[
            str,
            tuple[
                Dataset, Sequence[Example[PydanticSerializable, PydanticSerializable]]
            ],
        ] = {}

    def create_dataset(
        self,
        examples: Iterable[Example[Input, ExpectedOutput]],
        dataset_name: str,
        id: str | None = None,
        labels: set[str] | None = None,
        metadata: SerializableDict | None = None,
    ) -> Dataset:
        if metadata is None:
            metadata = dict()
        if labels is None:
            labels = set()
        dataset = Dataset(name=dataset_name, labels=labels, metadata=metadata)
        if id is not None:
            dataset.id = id
        if dataset.id in self._datasets_and_examples:
            if id:
                raise ValueError(
                    f"Cannot create dataset - dataset with given ID '{dataset.id}' already exists."
                )
            else:
                raise ValueError(
                    f"Newly assigned random dataset ID {dataset.id} already exists. This should not happen."
                )

        examples_casted = cast(
            Sequence[Example[PydanticSerializable, PydanticSerializable]],
            list(examples),
        )
        self._datasets_and_examples[dataset.id] = (dataset, examples_casted)

        return dataset

    def delete_dataset(self, dataset_id: str) -> None:
        self._datasets_and_examples.pop(dataset_id, None)

    def dataset(self, dataset_id: str) -> Optional[Dataset]:
        if dataset_id in self._datasets_and_examples:
            return self._datasets_and_examples[dataset_id][0]
        return None

    def dataset_ids(self) -> Iterable[str]:
        return sorted(list(self._datasets_and_examples.keys()))

    T = TypeVar("T")

    @staticmethod
    def _convert_to_type(data: Any, desired_type: T) -> T:
        if type(data) is desired_type:
            return data
        else:
            return TypeAdapter(desired_type).validate_python(data)

    def example(
        self,
        dataset_id: str,
        example_id: str,
        input_type: type[Input],
        expected_output_type: type[ExpectedOutput],
    ) -> Optional[Example[Input, ExpectedOutput]]:
        examples = self.examples(dataset_id, input_type, expected_output_type)
        filtered = (e for e in examples if e.id == example_id)
        return next(filtered, None)

    def examples(
        self,
        dataset_id: str,
        input_type: type[Input],
        expected_output_type: type[ExpectedOutput],
        examples_to_skip: Optional[frozenset[str]] = None,
    ) -> Iterable[Example[Input, ExpectedOutput]]:
        examples_to_skip = examples_to_skip or frozenset()
        if dataset_id not in self._datasets_and_examples:
            raise ValueError(
                f"Repository does not contain a dataset with id: {dataset_id}"
            )
        examples: list[Example[Input, ExpectedOutput]] = []
        for example in self._datasets_and_examples[dataset_id][1]:
            if example.id in examples_to_skip:
                continue
            converted_input = self._convert_to_type(example.input, input_type)
            converted_expected_output = self._convert_to_type(
                example.expected_output, expected_output_type
            )
            examples.append(
                Example[input_type, expected_output_type](  # type: ignore
                    id=example.id,
                    input=converted_input,
                    expected_output=converted_expected_output,
                    metadata=example.metadata,
                )
            )
        return sorted(
            examples,
            key=lambda example: example.id,
        )



from collections.abc import Sequence
from datetime import timedelta
from typing import cast

from aleph_alpha_client import CompletionResponse
from pharia_inference_sdk.core import ExportedSpan
from pharia_inference_sdk.core.model import _Complete
from pharia_inference_sdk.core.tracer.tracer import SpanType


def _get_root(trace: Sequence[ExportedSpan]) -> ExportedSpan | None:
    root_spans = [span for span in trace if span.parent_id is None]
    if len(root_spans) != 1:
        return None
    return root_spans[0]


def extract_latency_from_trace(trace: Sequence[ExportedSpan]) -> int:
    """Extract the total duration of a given trace based on its root trace.

    Args:
        trace: trace to analyze

    Returns:
        The duration of the trace in microseconds
    """
    root_span = _get_root(trace)
    if root_span is None:
        raise ValueError("No root span found in the trace")
    latency = (root_span.end_time - root_span.start_time) / timedelta(microseconds=1)
    return int(latency)


def _is_complete_request(span: ExportedSpan) -> bool:
    # Assuming that LLM requests have a specific name or attribute
    return span.name == _Complete.__name__


def _extract_tokens_from_complete_request(span: ExportedSpan) -> int:
    if not hasattr(span.attributes, "output"):
        raise ValueError(
            "Function expects a complete span with attributes.output. Output was not present."
        )
    completion_output = cast(CompletionResponse, span.attributes.output)
    return completion_output.num_tokens_generated


def extract_token_count_from_trace(trace: Sequence[ExportedSpan]) -> int:
    """Extract the number of tokens generated in a trace based on its completion requests.

    Note: Does not support traces of streamed responses.

    Args:
        trace: trace to analyze.

    Returns:
        The sum of newly generated tokens across all spans in the given trace.
    """
    token_count = 0
    for span in trace:
        if span.attributes.type != SpanType.TASK_SPAN:
            continue
        if _is_complete_request(span):
            token_count += _extract_tokens_from_complete_request(span)
    return token_count



from functools import _lru_cache_wrapper, lru_cache
from typing import Optional

from pharia_studio_sdk.evaluation.dataset.domain import Dataset
from pharia_studio_sdk.evaluation.dataset.file_dataset_repository import (
    FileSystemDatasetRepository,
)
from pharia_studio_sdk.evaluation.infrastructure.hugging_face_repository import (
    HuggingFaceRepository,
)


class HuggingFaceDatasetRepository(HuggingFaceRepository, FileSystemDatasetRepository):
    def __init__(
        self, repository_id: str, token: str, private: bool, caching: bool = True
    ) -> None:
        """Initializes a :class:`HuggingFaceDatasetRepository` to be ready for dataset storage and access.

        Args:
            repository_id: The HuggingFace namespace and repository name, separated by a "/".
            token: The HuggingFace authentication token.
            private: Whether the dataset repository should be private on HuggingFace.
            caching: If set, datasets are cached in memory once retrieved.
                This means external updates to datasets will be missed. Defaults to `True`.
        """
        super().__init__(repository_id, token, private)
        if caching:
            self.examples = lru_cache(maxsize=2)(self.examples)  # type: ignore

    def delete_dataset(self, dataset_id: str) -> None:
        """Deletes a dataset identified by the given dataset ID.

        This implementation should be backwards compatible to datasets
        created without a dataset object (i.e., there is no dataset file
        with dataset metadata).

        Note, that HuggingFace API does not seem to support deleting not-existing files.

        Args:
            dataset_id: Dataset ID of the dataset to delete.
        """
        if self.exists(self._dataset_examples_path(dataset_id)):
            self._file_system.rm(
                self.path_to_str(self._dataset_examples_path(dataset_id))
            )

        if self.exists(self._dataset_path(dataset_id)):
            self._file_system.rm(self.path_to_str(self._dataset_path(dataset_id)))
            # this resets the complete cache if a dataset gets deleted.
        if isinstance(self.examples, _lru_cache_wrapper):
            self.examples.cache_clear()

    def dataset(self, dataset_id: str) -> Optional[Dataset]:
        """Returns a dataset identified by the given dataset ID.

        This implementation should be backwards compatible to datasets
        created without a dataset object (i.e., there is no dataset file
        with dataset metadata).

        Args:
            dataset_id: Dataset ID of the dataset to delete.

        Returns:
            :class:`Dataset` if it was not, `None` otherwise.
        """
        dataset_file_path = self._dataset_path(dataset_id)
        examples_file_path = self._dataset_examples_path(dataset_id)
        if not self.exists(dataset_file_path):
            if not self.exists(examples_file_path):
                return None
            else:
                return Dataset(id=dataset_id, name=f"HuggingFace dataset {dataset_id}")

        return super().dataset(dataset_id)



import itertools
from collections.abc import Iterable, Sequence
from typing import Generic, Optional

import pandas as pd
import rich
from pharia_inference_sdk.core import Input, Output, Tracer
from rich.tree import Tree

from pharia_studio_sdk.evaluation.aggregation.domain import (
    AggregatedEvaluation,
    AggregationOverview,
)
from pharia_studio_sdk.evaluation.dataset.dataset_repository import DatasetRepository
from pharia_studio_sdk.evaluation.dataset.domain import Example, ExpectedOutput
from pharia_studio_sdk.evaluation.evaluation.domain import (
    Evaluation,
    ExampleEvaluation,
    FailedExampleEvaluation,
)
from pharia_studio_sdk.evaluation.evaluation.evaluation_repository import (
    EvaluationRepository,
)
from pharia_studio_sdk.evaluation.run.domain import ExampleOutput, FailedExampleRun
from pharia_studio_sdk.evaluation.run.run_repository import RunRepository

EXAMPLE_OUTPUT_TYPE = ExampleOutput[Output] | ExampleOutput[FailedExampleRun]


class RunLineage(Generic[Input, ExpectedOutput, Output]):
    example: Example[Input, ExpectedOutput]
    output: EXAMPLE_OUTPUT_TYPE
    tracer: Optional[Tracer]

    def __init__(
        self,
        example: Example[Input, ExpectedOutput],
        output: EXAMPLE_OUTPUT_TYPE,
        tracer: Optional[Tracer] = None,
    ) -> None:
        self.example = example
        self.output = output
        self.tracer = tracer

    def _rich_render(self) -> Tree:
        tree = Tree("Run Lineage")
        tree.add(self.example._rich_render())
        tree.add(self.output._rich_render(skip_example_id=True))
        return tree

    def _ipython_display_(self) -> None:
        rich.print(self._rich_render())


def run_lineages_to_pandas(
    run_lineages: Sequence[RunLineage[Input, ExpectedOutput, Output]],
) -> pd.DataFrame:
    """Converts a sequence of `RunLineage` objects to a pandas `DataFrame`.

    The `RunLineage` objects are stored in the column `"lineage"`.
    The `DataFrame` is indexed by `(example_id, run_id)`.

    Args:
        run_lineages: The lineages to convert.

    Returns:
        A pandas `DataFrame` with the data contained in the `run_lineages`.
    """
    df = pd.DataFrame(
        [
            vars(lineage.example) | vars(lineage.output) | {"lineage": lineage}
            for lineage in run_lineages
        ]
    )
    df = df.drop(columns="id")
    df = df.set_index(["example_id", "run_id"])
    return df


EXAMPLE_EVAL_TYPE = (
    ExampleEvaluation[Evaluation] | ExampleEvaluation[FailedExampleEvaluation]
)


class EvaluationLineage(Generic[Input, ExpectedOutput, Output, Evaluation]):
    example: Example[Input, ExpectedOutput]
    outputs: Sequence[EXAMPLE_OUTPUT_TYPE]
    evaluation: EXAMPLE_EVAL_TYPE
    tracers: Sequence[Optional[Tracer]]

    def __init__(
        self,
        example: Example[Input, ExpectedOutput],
        outputs: Sequence[EXAMPLE_OUTPUT_TYPE],
        evaluation: EXAMPLE_EVAL_TYPE,
        tracers: Sequence[Optional[Tracer]],
    ) -> None:
        self.example = example
        self.outputs = outputs
        self.evaluation = evaluation
        self.tracers = tracers

    def _rich_render(self) -> Tree:
        tree = Tree("Run Lineage")
        tree.add(self.example._rich_render())
        output_tree = Tree("Outputs")
        for output in self.outputs:
            output_tree.add(output._rich_render(skip_example_id=True))
        tree.add(output_tree)
        tree.add(self.evaluation._rich_render(skip_example_id=True))
        return tree

    def _ipython_display_(self) -> None:
        rich.print(self._rich_render())


def evaluation_lineages_to_pandas(
    evaluation_lineages: Sequence[
        EvaluationLineage[Input, ExpectedOutput, Output, Evaluation]
    ],
) -> pd.DataFrame:
    """Converts a sequence of `EvaluationLineage` objects to a pandas `DataFrame`.

    The `EvaluationLineage` objects are stored in the column `"lineage"`.
    The `DataFrame` is indexed by `(example_id, evaluation_id, run_id)`.
    Each `output` of every lineage will contribute one row in the `DataFrame`.

    Args:
        evaluation_lineages: The lineages to convert.

    Returns:
        A pandas `DataFrame` with the data contained in the `evaluation_lineages`.
    """
    df = pd.DataFrame(
        [
            vars(lineage.example)
            | vars(output)
            | vars(lineage.evaluation)
            | {"tracer": lineage.tracers[index]}
            | {"lineage": lineage}
            for lineage in evaluation_lineages
            for index, output in enumerate(lineage.outputs)
        ]
    )
    df = df.drop(columns="id")
    df = df.set_index(["example_id", "evaluation_id", "run_id"])
    return df


def aggregation_overviews_to_pandas(
    aggregation_overviews: Sequence[AggregationOverview[AggregatedEvaluation]],
    unwrap_statistics: bool = True,
    strict: bool = True,
    unwrap_metadata: bool = True,
) -> pd.DataFrame:
    """Converts aggregation overviews to a pandas table for easier comparison.

    Args:
        aggregation_overviews: Overviews to convert.
        unwrap_statistics: Unwrap the `statistics` field in the overviews into separate columns.
            Defaults to True.
        strict: Allow only overviews with exactly equal `statistics` types. Defaults to True.
        unwrap_metadata: Unwrap the `metadata` field in the overviews into separate columns.
            Defaults to True.

    Returns:
        A pandas :class:`DataFrame` containing an overview per row with fields as columns.
    """
    overviews = list(aggregation_overviews)
    if strict and len(overviews) > 1:
        first_type = overviews[0].statistics.__class__
        if any(
            overview.statistics.__class__ != first_type for overview in overviews[1:]
        ):
            raise ValueError(
                "Aggregation overviews contain different types, which is not allowed with strict=True"
            )

    df = pd.DataFrame(
        [model.model_dump(mode="json") for model in aggregation_overviews]
    )
    if unwrap_statistics and "statistics" in df.columns:
        df = df.join(pd.DataFrame(df["statistics"].to_list())).drop(
            columns=["statistics"]
        )
    if unwrap_metadata and "metadata" in df.columns:
        df = pd.concat([df, pd.json_normalize(df["metadata"])], axis=1).drop(  # type: ignore
            columns=["metadata"]
        )

    return df


class RepositoryNavigator:
    """The `RepositoryNavigator` is used to retrieve coupled data from multiple repositories."""

    def __init__(
        self,
        dataset_repository: DatasetRepository,
        run_repository: RunRepository,
        evaluation_repository: EvaluationRepository | None = None,
    ) -> None:
        self._dataset_repository = dataset_repository
        self._run_repository = run_repository
        self._eval_repository = evaluation_repository

    def run_lineages(
        self,
        run_id: str,
        input_type: type[Input],
        expected_output_type: type[ExpectedOutput],
        output_type: type[Output],
    ) -> Iterable[RunLineage[Input, ExpectedOutput, Output]]:
        """Retrieves all :class:`RunLineage`s for the run with id `run_id`.

        Args:
            run_id: The id of the run
            input_type: The type of the input as defined by the :class:`Example`
            expected_output_type: The type of the expected output as defined by the :class:`Example`
            output_type: The type of the run output as defined by the :class:`Output`

        Yields:
            An iterator over all :class:`RunLineage`s for the given run id.
        """
        run_overview = self._run_repository.run_overview(run_id)
        if run_overview is None:
            raise ValueError(f"Run repository does not contain a run with id {run_id}.")

        examples = list(
            self._dataset_repository.examples(
                run_overview.dataset_id,
                input_type,
                expected_output_type,
            )
        )

        example_outputs = list(
            self._run_repository.example_outputs(run_id, output_type)
        )

        # join
        for example, example_output in itertools.product(examples, example_outputs):
            if example.id == example_output.example_id:
                yield RunLineage(
                    example=example,
                    output=example_output,
                    tracer=self._run_repository.example_tracer(
                        run_id=run_id, example_id=example.id
                    ),
                )

    def evaluation_lineages(
        self,
        evaluation_id: str,
        input_type: type[Input],
        expected_output_type: type[ExpectedOutput],
        output_type: type[Output],
        evaluation_type: type[Evaluation],
    ) -> Iterable[EvaluationLineage[Input, ExpectedOutput, Output, Evaluation]]:
        """Retrieves all :class:`EvaluationLineage`s for the evaluation with id `evaluation_id`.

        Args:
            evaluation_id: The id of the evaluation
            input_type: The type of the input as defined by the :class:`Example`
            expected_output_type: The type of the expected output as defined by the :class:`Example`
            output_type: The type of the run output as defined by the :class:`Output`
            evaluation_type: The type of the evaluation as defined by the :class:`Evaluation`

        Yields:
            All :class:`EvaluationLineage`s for the given evaluation id.
        """
        if self._eval_repository is None:
            raise ValueError("Evaluation Repository is not set, but required.")

        eval_overview = self._eval_repository.evaluation_overview(evaluation_id)
        if eval_overview is None:
            raise ValueError(
                f"Evaluation repository does not contain an evaluation with id {evaluation_id}."
            )

        evaluations = list(
            self._eval_repository.example_evaluations(evaluation_id, evaluation_type)
        )
        run_lineages = list(
            itertools.chain.from_iterable(
                self.run_lineages(
                    overview.id, input_type, expected_output_type, output_type
                )
                for overview in eval_overview.run_overviews
            )
        )

        # join
        for evaluation in evaluations:
            example = None
            outputs = []
            tracers = []
            for run_lineage in run_lineages:
                if run_lineage.example.id == evaluation.example_id:
                    if example is None:
                        # the evaluation has only one example
                        # and all relevant run lineages contain the same example
                        example = run_lineage.example
                    outputs.append(run_lineage.output)
                    tracers.append(
                        self._run_repository.example_tracer(
                            run_lineage.output.run_id, run_lineage.output.example_id
                        )
                    )

            if example is not None:
                yield EvaluationLineage(
                    example=example,
                    outputs=outputs,
                    evaluation=evaluation,
                    tracers=tracers,
                )

    def run_lineage(
        self,
        run_id: str,
        example_id: str,
        input_type: type[Input],
        expected_output_type: type[ExpectedOutput],
        output_type: type[Output],
    ) -> RunLineage[Input, ExpectedOutput, Output] | None:
        """Retrieves the :class:`RunLineage` for the run with id `run_id` and example with id `example_id`.

        Args:
            run_id: The id of the run
            example_id: The id of the example
            input_type: The type of the input as defined by the :class:`Example`
            expected_output_type: The type of the expected output as defined by the :class:`Example`
            output_type: The type of the run output as defined by the :class:`Output`

        Returns:
            The :class:`RunLineage` for the given run id and example id, `None` if the example or an output for the example does not exist.
        """
        run_overview = self._run_repository.run_overview(run_id)
        if run_overview is None:
            raise ValueError(f"Run repository does not contain a run with id {run_id}.")

        example = self._dataset_repository.example(
            run_overview.dataset_id, example_id, input_type, expected_output_type
        )
        if example is None:
            return None

        example_output = self._run_repository.example_output(
            run_id, example_id, output_type
        )
        if example_output is None:
            return None

        return RunLineage(
            example=example,
            output=example_output,
            tracer=self._run_repository.example_tracer(run_id, example_id),
        )

    def evaluation_lineage(
        self,
        evaluation_id: str,
        example_id: str,
        input_type: type[Input],
        expected_output_type: type[ExpectedOutput],
        output_type: type[Output],
        evaluation_type: type[Evaluation],
    ) -> EvaluationLineage[Input, ExpectedOutput, Output, Evaluation] | None:
        """Retrieves the :class:`EvaluationLineage` for the evaluation with id `evaluation_id` and example with id `example_id`.

        Args:
            evaluation_id: The id of the evaluation
            example_id: The id of the example of interest
            input_type: The type of the input as defined by the :class:`Example`
            expected_output_type: The type of the expected output as defined by the :class:`Example`
            output_type: The type of the run output as defined by the :class:`Output`
            evaluation_type: The type of the evaluation as defined by the :class:`Evaluation`

        Returns:
            The :class:`EvaluationLineage` for the given evaluation id and example id.
            Returns `None` if the lineage is not complete because either an example, a run, or an evaluation does not exist.
        """
        if self._eval_repository is None:
            raise ValueError("Evaluation Repository is not set, but required.")

        eval_overview = self._eval_repository.evaluation_overview(evaluation_id)
        if eval_overview is None:
            raise ValueError(
                f"Evaluation repository does not contain an evaluation with id {evaluation_id}."
            )

        run_lineages = [
            self.run_lineage(
                overview.id, example_id, input_type, expected_output_type, output_type
            )
            for overview in eval_overview.run_overviews
        ]
        existing_run_lineages = [
            lineage for lineage in run_lineages if lineage is not None
        ]
        if len(existing_run_lineages) == 0:
            return None

        example_evaluation = self._eval_repository.example_evaluation(
            evaluation_id, example_id, evaluation_type
        )
        if example_evaluation is None:
            return None

        return EvaluationLineage(
            example=existing_run_lineages[0].example,
            outputs=[lineage.output for lineage in existing_run_lineages],
            evaluation=example_evaluation,
            tracers=[lineage.tracer for lineage in existing_run_lineages],
        )



from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING

from typing_extensions import TypeAliasType

if TYPE_CHECKING:
    JsonSerializable = (
        int
        | float
        | str
        | None
        | bool
        | Sequence["JsonSerializable"]
        | Mapping[str, "JsonSerializable"]
    )
else:
    JsonSerializable = TypeAliasType(
        "JsonSerializable",
        int
        | float
        | str
        | None
        | bool
        | Sequence["JsonSerializable"]
        | Mapping[str, "JsonSerializable"],
    )

SerializableDict = dict[str, JsonSerializable]



from collections import defaultdict
from collections.abc import Sequence
from typing import Optional

from pydantic import BaseModel

from pharia_studio_sdk.evaluation.evaluation.domain import (
    Evaluation,
    EvaluationOverview,
    ExampleEvaluation,
    FailedExampleEvaluation,
    PartialEvaluationOverview,
)
from pharia_studio_sdk.evaluation.evaluation.evaluation_repository import (
    EvaluationRepository,
)
from pharia_studio_sdk.evaluation.evaluation.evaluator.async_evaluator import (
    AsyncEvaluationRepository,
)


class InMemoryEvaluationRepository(EvaluationRepository):
    """An :class:`EvaluationRepository` that stores evaluation results in memory.

    Preferred for quick testing or to be used in Jupyter Notebooks.
    """

    def __init__(self) -> None:
        self._example_evaluations: dict[str, list[ExampleEvaluation[BaseModel]]] = (
            defaultdict(list)
        )
        self._evaluation_overviews: dict[str, EvaluationOverview] = dict()

    def store_evaluation_overview(self, overview: EvaluationOverview) -> None:
        self._evaluation_overviews[overview.id] = overview
        if overview.id not in self._example_evaluations:
            self._example_evaluations[overview.id] = []

    def evaluation_overview(self, evaluation_id: str) -> Optional[EvaluationOverview]:
        return self._evaluation_overviews.get(evaluation_id, None)

    def evaluation_overview_ids(self) -> Sequence[str]:
        return sorted(list(self._evaluation_overviews.keys()))

    def store_example_evaluation(
        self, evaluation: ExampleEvaluation[Evaluation]
    ) -> None:
        self._example_evaluations[evaluation.evaluation_id].append(evaluation)

    def _generate_evaluation_from_internal_evaluation(
        self,
        internal_evaluation: ExampleEvaluation[BaseModel],
        evaluation_type: type[Evaluation],
    ) -> ExampleEvaluation[Evaluation] | ExampleEvaluation[FailedExampleEvaluation]:
        if (
            type(internal_evaluation.result) is evaluation_type
            or type(internal_evaluation.result) is FailedExampleEvaluation
        ):
            return internal_evaluation  # type: ignore
        return ExampleEvaluation[Evaluation](
            evaluation_id=internal_evaluation.evaluation_id,
            example_id=internal_evaluation.example_id,
            result=evaluation_type.model_validate(internal_evaluation.result),
        )

    def example_evaluation(
        self, evaluation_id: str, example_id: str, evaluation_type: type[Evaluation]
    ) -> Optional[
        ExampleEvaluation[Evaluation] | ExampleEvaluation[FailedExampleEvaluation]
    ]:
        results = self.example_evaluations(evaluation_id, evaluation_type)
        filtered = (result for result in results if result.example_id == example_id)
        return next(filtered, None)

    def example_evaluations(
        self, evaluation_id: str, evaluation_type: type[Evaluation]
    ) -> Sequence[
        ExampleEvaluation[Evaluation] | ExampleEvaluation[FailedExampleEvaluation]
    ]:
        if evaluation_id not in self._example_evaluations:
            raise ValueError(
                f"Repository does not contain an evaluation with id: {evaluation_id}"
            )

        example_evaluations = [
            self._generate_evaluation_from_internal_evaluation(
                example_evaluation, evaluation_type
            )
            for example_evaluation in self._example_evaluations[evaluation_id]
        ]
        return sorted(example_evaluations, key=lambda i: i.example_id)


class AsyncInMemoryEvaluationRepository(
    AsyncEvaluationRepository, InMemoryEvaluationRepository
):
    def __init__(self) -> None:
        super().__init__()
        self._partial_evaluation_overviews: dict[str, PartialEvaluationOverview] = (
            dict()
        )

    def store_partial_evaluation_overview(
        self, overview: PartialEvaluationOverview
    ) -> None:
        self._partial_evaluation_overviews[overview.id] = overview
        if overview.id not in self._example_evaluations:
            self._example_evaluations[overview.id] = []

    def partial_evaluation_overview(
        self, evaluation_id: str
    ) -> Optional[PartialEvaluationOverview]:
        return self._partial_evaluation_overviews.get(evaluation_id, None)

    def partial_evaluation_overview_ids(self) -> Sequence[str]:
        return sorted(list(self._partial_evaluation_overviews.keys()))



import itertools
from collections.abc import Sequence
from datetime import datetime, timezone
from http import HTTPStatus
from typing import Any, Optional

import requests
from pharia_inference_sdk.core import Input, Output
from pharia_inference_sdk.core.task import Task
from pharia_inference_sdk.core.tracer.tracer import ExportedSpan
from pydantic import TypeAdapter
from tqdm import tqdm

from pharia_studio_sdk.connectors.studio.studio import (
    AggregationLogicIdentifier,
    BenchmarkLineage,
    EvaluationLogicIdentifier,
    PostBenchmarkExecution,
    StudioClient,
)
from pharia_studio_sdk.evaluation.aggregation.aggregator import (
    AggregationLogic,
    Aggregator,
)
from pharia_studio_sdk.evaluation.aggregation.domain import AggregatedEvaluation
from pharia_studio_sdk.evaluation.aggregation.in_memory_aggregation_repository import (
    InMemoryAggregationRepository,
)
from pharia_studio_sdk.evaluation.benchmark.benchmark import (
    Benchmark,
    BenchmarkRepository,
)
from pharia_studio_sdk.evaluation.benchmark.get_code import get_source_notebook_safe
from pharia_studio_sdk.evaluation.benchmark.trace_information import (
    extract_latency_from_trace,
    extract_token_count_from_trace,
)
from pharia_studio_sdk.evaluation.dataset.domain import ExpectedOutput
from pharia_studio_sdk.evaluation.dataset.studio_dataset_repository import (
    StudioDatasetRepository,
)
from pharia_studio_sdk.evaluation.evaluation.domain import Evaluation
from pharia_studio_sdk.evaluation.evaluation.evaluator.evaluator import (
    EvaluationLogic,
    Evaluator,
)
from pharia_studio_sdk.evaluation.evaluation.in_memory_evaluation_repository import (
    InMemoryEvaluationRepository,
)
from pharia_studio_sdk.evaluation.infrastructure.repository_navigator import (
    EvaluationLineage,
)
from pharia_studio_sdk.evaluation.run.in_memory_run_repository import (
    InMemoryRunRepository,
)
from pharia_studio_sdk.evaluation.run.runner import Runner


class StudioBenchmark(Benchmark):
    def __init__(
        self,
        benchmark_id: str,
        name: str,
        dataset_id: str,
        eval_logic: EvaluationLogic[Input, Output, ExpectedOutput, Evaluation],
        aggregation_logic: AggregationLogic[Evaluation, AggregatedEvaluation],
        studio_client: StudioClient,
        **kwargs: Any,
    ):
        self.id = benchmark_id
        self.name = name
        self.dataset_id = dataset_id
        self.eval_logic = eval_logic
        self.aggregation_logic = aggregation_logic
        self.client = studio_client
        self.run_repository = InMemoryRunRepository()
        self.evaluation_repository = InMemoryEvaluationRepository()
        self.aggregation_repository = InMemoryAggregationRepository()
        self.dataset_repository = StudioDatasetRepository(self.client)
        self.evaluator = Evaluator(
            self.dataset_repository,
            self.run_repository,
            self.evaluation_repository,
            f"benchmark-{self.id}-evaluator",
            self.eval_logic,
        )
        self.aggregator = Aggregator(
            self.evaluation_repository,
            self.aggregation_repository,
            f"benchmark-{self.id}-aggregator",
            self.aggregation_logic,
        )

    def execute(
        self,
        task: Task[Input, Output],
        name: str,
        description: Optional[str] = None,
        labels: Optional[set[str]] = None,
        metadata: Optional[dict[str, Any]] = None,
    ) -> str:
        start = datetime.now(timezone.utc)

        runner = Runner(
            task,
            self.dataset_repository,
            self.run_repository,
            f"benchmark-{self.id}-runner",
        )
        run_overview = runner.run_dataset(
            self.dataset_id, description=description, labels=labels, metadata=metadata
        )

        evaluation_overview = self.evaluator.evaluate_runs(
            run_overview.id, description=description, labels=labels, metadata=metadata
        )

        aggregation_overview = self.aggregator.aggregate_evaluation(
            evaluation_overview.id,
            description=description,
            labels=labels,
            metadata=metadata,
        )

        end = datetime.now(timezone.utc)

        evaluation_lineages = list(
            self.evaluator.evaluation_lineages(evaluation_overview.id)
        )

        run_traces = [
            self._trace_from_lineage(lineage) for lineage in evaluation_lineages
        ]
        tokens_per_trace = [
            extract_token_count_from_trace(trace) for trace in run_traces
        ]
        latency_per_trace = [extract_latency_from_trace(trace) for trace in run_traces]

        tokens_per_successful_trace, latency_per_successful_trace = (
            self._filter_for_succesful_runs(
                (tokens_per_trace, latency_per_trace),
                source_lineage_list=evaluation_lineages,
                run_id=run_overview.id,
            )
        )

        def average_or_zero(list: list) -> float:
            return sum(list) / len(list) if len(list) > 0 else 0

        benchmark_execution_data = PostBenchmarkExecution(
            name=name,
            description=description,
            labels=labels,
            metadata=metadata,
            start=start,
            end=end,
            run_start=run_overview.start,
            run_end=run_overview.end,
            run_successful_count=run_overview.successful_example_count,
            run_failed_count=run_overview.failed_example_count,
            run_success_avg_latency=average_or_zero(latency_per_successful_trace),
            run_success_avg_token_count=average_or_zero(tokens_per_successful_trace),
            eval_start=evaluation_overview.start_date,
            eval_end=evaluation_overview.end_date,
            eval_successful_count=evaluation_overview.successful_evaluation_count,
            eval_failed_count=evaluation_overview.failed_evaluation_count,
            aggregation_start=aggregation_overview.start,
            aggregation_end=aggregation_overview.end,
            statistics=aggregation_overview.statistics.model_dump_json(),
        )

        benchmark_execution_id = self.client.submit_benchmark_execution(
            benchmark_id=self.id, data=benchmark_execution_data
        )

        trace_ids = []
        for trace in tqdm(run_traces, desc="Submitting traces to Studio"):
            trace_id = self.client.submit_trace(trace)
            trace_ids.append(trace_id)

        benchmark_lineages = self._create_benchmark_lineages(
            eval_lineages=evaluation_lineages,
            trace_ids=trace_ids,
            latencies_per_trace=latency_per_trace,
            tokens_per_trace=tokens_per_trace,
        )

        self.client.submit_benchmark_lineages(
            benchmark_lineages=benchmark_lineages,
            execution_id=benchmark_execution_id,
            benchmark_id=self.id,
        )

        return benchmark_execution_id

    def _filter_for_succesful_runs(
        self,
        lists_to_filter: tuple[list, ...],
        source_lineage_list: list[
            EvaluationLineage[Input, ExpectedOutput, Output, Evaluation]
        ],
        run_id: str,
    ) -> tuple[list, ...]:
        """This method assumes that lists_to_filter and source_lineage_list are all equal length."""
        failed_example_output_ids = [
            example_output.example_id
            for example_output in self.run_repository.failed_example_outputs(
                run_id=run_id, output_type=self.evaluator.output_type()
            )
        ]

        is_successful_run = [
            lineage.example.id not in failed_example_output_ids
            for lineage in source_lineage_list
        ]
        return tuple(
            list(itertools.compress(sublist, is_successful_run))
            for sublist in lists_to_filter
        )

    def _trace_from_lineage(
        self, eval_lineage: EvaluationLineage[Input, ExpectedOutput, Output, Evaluation]
    ) -> Sequence[ExportedSpan]:
        # since we have 1 output only in this scenario, we expected to have exactly 1 tracer
        trace = eval_lineage.tracers[0]
        assert trace, "eval lineages always should have at least 1 tracer"
        return trace.export_for_viewing()

    def _create_benchmark_lineages(
        self,
        eval_lineages: list[
            EvaluationLineage[Input, ExpectedOutput, Output, Evaluation]
        ],
        trace_ids: list[str],
        latencies_per_trace: list[int],
        tokens_per_trace: list[int],
    ) -> Sequence[BenchmarkLineage[Input, ExpectedOutput, Output, Evaluation]]:
        return [
            self._create_benchmark_lineage(
                eval_lineage, trace_id, run_latency, run_tokens
            )
            for eval_lineage, trace_id, run_latency, run_tokens in zip(
                eval_lineages,
                trace_ids,
                latencies_per_trace,
                tokens_per_trace,
                strict=True,
            )
        ]

    def _create_benchmark_lineage(
        self,
        eval_lineage: EvaluationLineage[Input, ExpectedOutput, Output, Evaluation],
        trace_id: str,
        run_latency: int,
        run_tokens: int,
    ) -> BenchmarkLineage:
        return BenchmarkLineage(
            trace_id=trace_id,
            input=eval_lineage.example.input,
            expected_output=eval_lineage.example.expected_output,
            example_metadata=eval_lineage.example.metadata,
            output=eval_lineage.outputs[0].output,
            evaluation=eval_lineage.evaluation.result,
            run_latency=run_latency,
            run_tokens=run_tokens,
        )


class StudioBenchmarkRepository(BenchmarkRepository):
    def __init__(self, studio_client: StudioClient):
        self.client = studio_client

    def create_benchmark(
        self,
        dataset_id: str,
        eval_logic: EvaluationLogic[Input, Output, ExpectedOutput, Evaluation],
        aggregation_logic: AggregationLogic[Evaluation, AggregatedEvaluation],
        name: str,
        metadata: Optional[dict[str, Any]] = None,
        description: Optional[str] = None,
    ) -> StudioBenchmark:
        try:
            benchmark_id = self.client.submit_benchmark(
                dataset_id,
                create_evaluation_logic_identifier(eval_logic),
                create_aggregation_logic_identifier(aggregation_logic),
                name,
                description,
                metadata,
            )
        except requests.HTTPError as e:
            if (
                e.response.status_code == HTTPStatus.BAD_REQUEST
                or e.response.status_code == HTTPStatus.NOT_FOUND
            ):
                raise ValueError(f"Dataset with ID {dataset_id} not found") from e
            if e.response.status_code == HTTPStatus.CONFLICT:
                raise ValueError(
                    f"""Benchmark with name "{name}" already exists. Names of Benchmarks in the same Project must be unique."""
                ) from e
            else:
                raise ValueError(
                    "An error occurred when attempting to create a benchmark."
                ) from e

        return StudioBenchmark(
            benchmark_id,
            name,
            dataset_id,
            eval_logic,
            aggregation_logic,
            studio_client=self.client,
        )

    def get_benchmark(
        self,
        benchmark_id: str,
        eval_logic: EvaluationLogic[Input, Output, ExpectedOutput, Evaluation],
        aggregation_logic: AggregationLogic[Evaluation, AggregatedEvaluation],
        allow_diff: bool = False,
    ) -> StudioBenchmark | None:
        benchmark = self.client.get_benchmark(benchmark_id)
        if benchmark is None:
            return None
        return StudioBenchmark(
            benchmark_id,
            benchmark.name,
            benchmark.dataset_id,
            eval_logic,
            aggregation_logic,
            self.client,
        )


def type_to_schema(type_: type) -> dict[str, Any]:
    return TypeAdapter(type_).json_schema()


def create_evaluation_logic_identifier(
    eval_logic: EvaluationLogic[Input, Output, ExpectedOutput, Evaluation],
) -> EvaluationLogicIdentifier:
    evaluator = Evaluator(
        dataset_repository=None,  # type: ignore
        run_repository=None,  # type: ignore
        evaluation_repository=None,  # type: ignore
        description="",
        evaluation_logic=eval_logic,
    )
    return EvaluationLogicIdentifier(
        logic=get_source_notebook_safe(eval_logic),
        input_schema=type_to_schema(evaluator.input_type()),
        output_schema=type_to_schema(evaluator.output_type()),
        expected_output_schema=type_to_schema(evaluator.expected_output_type()),
        evaluation_schema=type_to_schema(evaluator.evaluation_type()),
    )


def create_aggregation_logic_identifier(
    aggregation_logic: AggregationLogic[Evaluation, AggregatedEvaluation],
) -> AggregationLogicIdentifier:
    aggregator = Aggregator(
        evaluation_repository=None,  # type: ignore
        aggregation_repository=None,  # type: ignore
        description="",
        aggregation_logic=aggregation_logic,
    )
    return AggregationLogicIdentifier(
        logic=get_source_notebook_safe(aggregation_logic),
        evaluation_schema=type_to_schema(aggregator.evaluation_type()),
        aggregation_schema=type_to_schema(aggregator.aggregated_evaluation_type()),
    )



from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Iterator, Mapping
from functools import cached_property
from typing import (
    Generic,
    Optional,
    TypeVar,
    cast,
    final,
    get_args,
    get_origin,
)
from uuid import uuid4

from pharia_inference_sdk.core import utc_now

from pharia_studio_sdk.connectors.base.json_serializable import (
    SerializableDict,
)
from pharia_studio_sdk.evaluation.aggregation.aggregation_repository import (
    AggregationRepository,
)
from pharia_studio_sdk.evaluation.aggregation.domain import (
    AggregatedEvaluation,
    AggregationOverview,
)
from pharia_studio_sdk.evaluation.evaluation.domain import (
    Evaluation,
    EvaluationOverview,
    FailedExampleEvaluation,
)
from pharia_studio_sdk.evaluation.evaluation.evaluation_repository import (
    EvaluationRepository,
)

T = TypeVar("T")


class CountingFilterIterable(Iterable[T]):
    def __init__(
        self, wrapped_iterable: Iterable[T], filter: Callable[[T], bool]
    ) -> None:
        self._wrapped_iterator = iter(wrapped_iterable)
        self._filter = filter
        self._included_count = 0
        self._excluded_count = 0

    def __next__(self) -> T:
        e = next(self._wrapped_iterator)
        while not self._filter(e):
            self._excluded_count += 1
            e = next(self._wrapped_iterator)
        self._included_count += 1
        return e

    def __iter__(self) -> Iterator[T]:
        return self

    def included_count(self) -> int:
        return self._included_count

    def excluded_count(self) -> int:
        return self._excluded_count


class AggregationLogic(ABC, Generic[Evaluation, AggregatedEvaluation]):
    @abstractmethod
    def aggregate(self, evaluations: Iterable[Evaluation]) -> AggregatedEvaluation:
        """`Evaluator`-specific method for aggregating individual `Evaluations` into report-like `Aggregated Evaluation`.

        This method is responsible for taking the results of an evaluation run and aggregating all the results.
        It should create an `AggregatedEvaluation` class and return it at the end.

        Args:
            evaluations: The results from running `eval_and_aggregate_runs` with a :class:`Task`.

        Returns:
            The aggregated results of an evaluation run with a :class:`Dataset`.
        """
        ...


class Aggregator(Generic[Evaluation, AggregatedEvaluation]):
    """Aggregator that can handle automatic aggregation of evaluation scenarios.

    This aggregator should be used for automatic eval. A user still has to implement
    :class: `AggregationLogic`.


    Arguments:
        evaluation_repository: The repository that will be used to store evaluation results.
        aggregation_repository: The repository that will be used to store aggregation results.
        description: Human-readable description for the evaluator.
        aggregation_logic: The logic to aggregate the evaluations.

    Generics:
        Evaluation: Interface of the metrics that come from the evaluated :class:`Task`.
        AggregatedEvaluation: The aggregated results of an evaluation run with a :class:`Dataset`.
    """

    def __init__(
        self,
        evaluation_repository: EvaluationRepository,
        aggregation_repository: AggregationRepository,
        description: str,
        aggregation_logic: AggregationLogic[Evaluation, AggregatedEvaluation],
    ) -> None:
        self._evaluation_repository = evaluation_repository
        self._aggregation_repository = aggregation_repository
        self._aggregation_logic = aggregation_logic
        self.description = description

    @cached_property
    def _get_types(self) -> Mapping[str, type]:
        """Type magic function that gets the actual types of the generic parameters.

        Traverses the inheritance history of `AggregationLogic`-subclass to find an actual type every time a TypeVar is replaced.

        Returns:
            Name of generic parameter to the type found.
        """

        def is_eligible_subclass(parent: type) -> bool:
            return hasattr(parent, "__orig_bases__") and issubclass(
                parent, AggregationLogic
            )

        def update_types() -> None:
            num_types_set = 0
            for current_index, current_type in enumerate(current_types):
                if type(current_type) is not TypeVar:
                    type_var_count = num_types_set - 1
                    final_index = -1
                    for element_index, element in enumerate(type_list):
                        final_index = element_index
                        if type(element) is TypeVar:
                            type_var_count += 1
                        if type_var_count == current_index:
                            break
                    assert type_var_count == current_index
                    type_list[final_index] = current_type
                    num_types_set += 1

        # mypy does not know __orig_bases__
        base_types = AggregationLogic.__orig_bases__[1]  # type: ignore
        type_list: list[type | TypeVar] = list(get_args(base_types))

        possible_parent_classes = [
            p
            for p in reversed(type(self._aggregation_logic).__mro__)
            if is_eligible_subclass(p)
        ]
        for parent in possible_parent_classes:
            # mypy does not know __orig_bases__
            for base in parent.__orig_bases__:  # type: ignore
                origin = get_origin(base)
                if origin is None or not issubclass(origin, AggregationLogic):
                    continue
                current_types = list(get_args(base))
                update_types()

        return {
            name: param_type
            for name, param_type in zip(
                (a.__name__ for a in get_args(base_types)), type_list, strict=False
            )
            if type(param_type) is not TypeVar
        }

    def evaluation_type(self) -> type[Evaluation]:
        """Returns the type of the evaluation result of an example.

        This can be used to retrieve properly typed evaluations of an evaluation run
        from a :class:`EvaluationRepository`

        Returns:
            Returns the type of the evaluation result of an example.
        """
        try:
            evaluation_type = self._get_types["Evaluation"]
        except KeyError:
            raise TypeError(
                f"Alternatively overwrite evaluation_type() in {type(self)}"
            ) from None
        return cast(type[Evaluation], evaluation_type)

    def aggregated_evaluation_type(self) -> type[AggregatedEvaluation]:
        """Returns the type of the aggregated result of a run.

        Returns:
            Returns the type of the aggreagtion result.
        """
        try:
            aggregated_evaluation_type = self._get_types["AggregatedEvaluation"]
        except KeyError:
            raise TypeError(
                f"Alternatively overwrite aggregated_evaluation_type() in {type(self)}"
            ) from None
        return cast(type[AggregatedEvaluation], aggregated_evaluation_type)

    @final
    def aggregate_evaluation(
        self,
        *eval_ids: str,
        description: Optional[str] = None,
        labels: set[str] | None = None,
        metadata: SerializableDict | None = None,
    ) -> AggregationOverview[AggregatedEvaluation]:
        """Aggregates all evaluations into an overview that includes high-level statistics.

        Aggregates :class:`Evaluation`s according to the implementation of :func:`AggregationLogic.aggregate`.

        Args:
            *eval_ids: An overview of the evaluation to be aggregated. Does not include
                actual evaluations as these will be retrieved from the repository.
            description: Optional description of the aggregation. Defaults to None.
            labels: A list of labels for filtering. Defaults to an empty list.
            metadata: A dict for additional information about the aggregation overview. Defaults to an empty dict.

        Returns:
            An overview of the aggregated evaluation.
        """
        if metadata is None:
            metadata = dict()
        if labels is None:
            labels = set()

        def load_eval_overview(evaluation_id: str) -> EvaluationOverview:
            evaluation_overview = self._evaluation_repository.evaluation_overview(
                evaluation_id
            )
            if not evaluation_overview:
                raise ValueError(
                    f"No PartialEvaluationOverview found for eval-id: {evaluation_id}"
                )
            return evaluation_overview

        evaluation_overviews = frozenset(
            load_eval_overview(evaluation_id) for evaluation_id in set(eval_ids)
        )

        nested_evaluations = [
            self._evaluation_repository.example_evaluations(
                overview.id, self.evaluation_type()
            )
            for overview in evaluation_overviews
        ]
        example_evaluations = [
            evaluation for sublist in nested_evaluations for evaluation in sublist
        ]

        successful_evaluations = CountingFilterIterable(
            (example_eval.result for example_eval in example_evaluations),
            lambda evaluation: not isinstance(evaluation, FailedExampleEvaluation),
        )
        start = utc_now()
        statistics = self._aggregation_logic.aggregate(
            cast(Iterable[Evaluation], successful_evaluations)
        )

        full_description = (
            self.description + " : " + description if description else self.description
        )
        aggregation_overview = AggregationOverview(
            evaluation_overviews=frozenset(evaluation_overviews),
            id=str(uuid4()),
            start=start,
            end=utc_now(),
            successful_evaluation_count=successful_evaluations.included_count(),
            crashed_during_evaluation_count=successful_evaluations.excluded_count(),
            description=full_description,
            statistics=statistics,
            labels=labels,
            metadata=metadata,
        )
        self._aggregation_repository.store_aggregation_overview(aggregation_overview)
        return aggregation_overview



from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from typing import Optional

from pharia_inference_sdk.core.task import Input, Output

from pharia_studio_sdk.evaluation.dataset.domain import ExpectedOutput
from pharia_studio_sdk.evaluation.evaluation.domain import (
    Evaluation,
    EvaluationOverview,
    PartialEvaluationOverview,
)
from pharia_studio_sdk.evaluation.evaluation.evaluation_repository import (
    EvaluationRepository,
)
from pharia_studio_sdk.evaluation.evaluation.evaluator.base_evaluator import (
    EvaluatorBase,
)


class AsyncEvaluator(EvaluatorBase[Input, Output, ExpectedOutput, Evaluation], ABC):
    @abstractmethod
    def submit(
        self,
        *run_ids: str,
        num_examples: Optional[int] = None,
        abort_on_error: bool = False,
    ) -> PartialEvaluationOverview:
        """Submits evaluations to external service to be evaluated.

        Failed submissions are saved as FailedExampleEvaluations.

        Args:
            *run_ids: The runs to be evaluated. Each run is expected to have the same
                dataset as input (which implies their tasks have the same input-type)
                and their tasks have the same output-type. For each example in the
                dataset referenced by the runs the outputs of all runs are collected
                and if all of them were successful they are passed on to the implementation
                specific evaluation. The method compares all run of the provided ids to each other.
            num_examples: The number of examples which should be evaluated from the given runs.
                Always the first n runs stored in the evaluation repository. Defaults to None.
            abort_on_error: Abort the whole submission process if a single submission fails.
                Defaults to False.

        Returns:
            A :class:`PartialEvaluationOverview` containing submission information.
        """
        ...

    @abstractmethod
    def retrieve(self, partial_overview_id: str) -> EvaluationOverview:
        """Retrieves external evaluations and saves them to an evaluation repository.

        Failed or skipped submissions should be viewed as failed evaluations.
        Evaluations that are submitted but not yet evaluated also count as failed evaluations.

        Args:
            partial_overview_id: The id of the corresponding :class:`PartialEvaluationOverview`.

        Returns:
            An :class:`EvaluationOverview` that describes the whole evaluation.
        """
        ...


class AsyncEvaluationRepository(EvaluationRepository):
    @abstractmethod
    def store_partial_evaluation_overview(
        self, partial_evaluation_overview: PartialEvaluationOverview
    ) -> None:
        """Stores an :class:`PartialEvaluationOverview`.

        Args:
            partial_evaluation_overview: The partial overview to be persisted.
        """
        ...

    @abstractmethod
    def partial_evaluation_overview(
        self, partial_evaluation_id: str
    ) -> Optional[PartialEvaluationOverview]:
        """Returns an :class:`PartialEvaluationOverview` for the given ID.

        Args:
            partial_evaluation_id: ID of the partial evaluation overview to retrieve.

        Returns:
            :class:`PartialEvaluationOverview` if it was found, `None` otherwise.
        """
        ...

    def partial_evaluation_overviews(self) -> Iterable[PartialEvaluationOverview]:
        """Returns all :class:`PartialEvaluationOverview`s sorted by their ID.

        Yields:
            :class:`PartialEvaluationOverview`s.
        """
        for eval_id in self.partial_evaluation_overview_ids():
            evaluation_overview = self.partial_evaluation_overview(eval_id)
            if evaluation_overview is not None:
                yield evaluation_overview

    @abstractmethod
    def partial_evaluation_overview_ids(self) -> Sequence[str]:
        """Returns sorted IDs of all stored :class:`PartialEvaluationOverview`s.

        Returns:
            A :class:`Sequence` of the :class:`PartialEvaluationOverview` IDs.
        """
        ...



from pharia_studio_sdk.connectors.argilla.argilla_client import (
    ArgillaClient,
    ArgillaEvaluation,
    Record,
    RecordData,
)
from pharia_studio_sdk.connectors.base.json_serializable import (
    JsonSerializable,
    SerializableDict,
)
from pharia_studio_sdk.connectors.studio.studio import (
    AggregationLogicIdentifier,
    BenchmarkLineage,
    EvaluationLogicIdentifier,
    GetBenchmarkLineageResponse,
    GetBenchmarkResponse,
    GetDatasetExamplesResponse,
    PostBenchmarkExecution,
    PostBenchmarkLineagesRequest,
    PostBenchmarkLineagesResponse,
    PostBenchmarkRequest,
    StudioClient,
    StudioDataset,
    StudioExample,
    StudioProject,
)

__all__ = [
    "AggregationLogicIdentifier",
    # Argilla components
    "ArgillaClient",
    "ArgillaEvaluation",
    "BenchmarkLineage",
    "EvaluationLogicIdentifier",
    "GetBenchmarkLineageResponse",
    "GetBenchmarkResponse",
    "GetDatasetExamplesResponse",
    # Base types
    "JsonSerializable",
    "PostBenchmarkExecution",
    "PostBenchmarkLineagesRequest",
    "PostBenchmarkLineagesResponse",
    "PostBenchmarkRequest",
    "Record",
    "RecordData",
    "SerializableDict",
    # Studio components
    "StudioClient",
    "StudioDataset",
    "StudioExample",
    "StudioProject",
]



"""
The following code is licensed under the following license and taken from https://github.com/wandb/weave.
The following modifications have been made:
 - Adapt the type hints to be more specific to this project
 - Change the control flow for the get_source_notebook_safe to try to only get the source from notebooks once it fails.

                              Apache License
                        Version 2.0, January 2004
                     http://www.apache.org/licenses/

TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION

1. Definitions.

   "License" shall mean the terms and conditions for use, reproduction,
   and distribution as defined by Sections 1 through 9 of this document.

   "Licensor" shall mean the copyright owner or entity authorized by
   the copyright owner that is granting the License.

   "Legal Entity" shall mean the union of the acting entity and all
   other entities that control, are controlled by, or are under common
   control with that entity. For the purposes of this definition,
   "control" means (i) the power, direct or indirect, to cause the
   direction or management of such entity, whether by contract or
   otherwise, or (ii) ownership of fifty percent (50%) or more of the
   outstanding shares, or (iii) beneficial ownership of such entity.

   "You" (or "Your") shall mean an individual or Legal Entity
   exercising permissions granted by this License.

   "Source" form shall mean the preferred form for making modifications,
   including but not limited to software source code, documentation
   source, and configuration files.

   "Object" form shall mean any form resulting from mechanical
   transformation or translation of a Source form, including but
   not limited to compiled object code, generated documentation,
   and conversions to other media types.

   "Work" shall mean the work of authorship, whether in Source or
   Object form, made available under the License, as indicated by a
   copyright notice that is included in or attached to the work
   (an example is provided in the Appendix below).

   "Derivative Works" shall mean any work, whether in Source or Object
   form, that is based on (or derived from) the Work and for which the
   editorial revisions, annotations, elaborations, or other modifications
   represent, as a whole, an original work of authorship. For the purposes
   of this License, Derivative Works shall not include works that remain
   separable from, or merely link (or bind by name) to the interfaces of,
   the Work and Derivative Works thereof.

   "Contribution" shall mean any work of authorship, including
   the original version of the Work and any modifications or additions
   to that Work or Derivative Works thereof, that is intentionally
   submitted to Licensor for inclusion in the Work by the copyright owner
   or by an individual or Legal Entity authorized to submit on behalf of
   the copyright owner. For the purposes of this definition, "submitted"
   means any form of electronic, verbal, or written communication sent
   to the Licensor or its representatives, including but not limited to
   communication on electronic mailing lists, source code control systems,
   and issue tracking systems that are managed by, or on behalf of, the
   Licensor for the purpose of discussing and improving the Work, but
   excluding communication that is conspicuously marked or otherwise
   designated in writing by the copyright owner as "Not a Contribution."

   "Contributor" shall mean Licensor and any individual or Legal Entity
   on behalf of whom a Contribution has been received by Licensor and
   subsequently incorporated within the Work.

2. Grant of Copyright License. Subject to the terms and conditions of
   this License, each Contributor hereby grants to You a perpetual,
   worldwide, non-exclusive, no-charge, royalty-free, irrevocable
   copyright license to reproduce, prepare Derivative Works of,
   publicly display, publicly perform, sublicense, and distribute the
   Work and such Derivative Works in Source or Object form.

3. Grant of Patent License. Subject to the terms and conditions of
   this License, each Contributor hereby grants to You a perpetual,
   worldwide, non-exclusive, no-charge, royalty-free, irrevocable
   (except as stated in this section) patent license to make, have made,
   use, offer to sell, sell, import, and otherwise transfer the Work,
   where such license applies only to those patent claims licensable
   by such Contributor that are necessarily infringed by their
   Contribution(s) alone or by combination of their Contribution(s)
   with the Work to which such Contribution(s) was submitted. If You
   institute patent litigation against any entity (including a
   cross-claim or counterclaim in a lawsuit) alleging that the Work
   or a Contribution incorporated within the Work constitutes direct
   or contributory patent infringement, then any patent licenses
   granted to You under this License for that Work shall terminate
   as of the date such litigation is filed.

4. Redistribution. You may reproduce and distribute copies of the
   Work or Derivative Works thereof in any medium, with or without
   modifications, and in Source or Object form, provided that You
   meet the following conditions:

   (a) You must give any other recipients of the Work or
       Derivative Works a copy of this License; and

   (b) You must cause any modified files to carry prominent notices
       stating that You changed the files; and

   (c) You must retain, in the Source form of any Derivative Works
       that You distribute, all copyright, patent, trademark, and
       attribution notices from the Source form of the Work,
       excluding those notices that do not pertain to any part of
       the Derivative Works; and

   (d) If the Work includes a "NOTICE" text file as part of its
       distribution, then any Derivative Works that You distribute must
       include a readable copy of the attribution notices contained
       within such NOTICE file, excluding those notices that do not
       pertain to any part of the Derivative Works, in at least one
       of the following places: within a NOTICE text file distributed
       as part of the Derivative Works; within the Source form or
       documentation, if provided along with the Derivative Works; or,
       within a display generated by the Derivative Works, if and
       wherever such third-party notices normally appear. The contents
       of the NOTICE file are for informational purposes only and
       do not modify the License. You may add Your own attribution
       notices within Derivative Works that You distribute, alongside
       or as an addendum to the NOTICE text from the Work, provided
       that such additional attribution notices cannot be construed
       as modifying the License.

   You may add Your own copyright statement to Your modifications and
   may provide additional or different license terms and conditions
   for use, reproduction, or distribution of Your modifications, or
   for any such Derivative Works as a whole, provided Your use,
   reproduction, and distribution of the Work otherwise complies with
   the conditions stated in this License.

5. Submission of Contributions. Unless You explicitly state otherwise,
   any Contribution intentionally submitted for inclusion in the Work
   by You to the Licensor shall be under the terms and conditions of
   this License, without any additional terms or conditions.
   Notwithstanding the above, nothing herein shall supersede or modify
   the terms of any separate license agreement you may have executed
   with Licensor regarding such Contributions.

6. Trademarks. This License does not grant permission to use the trade
   names, trademarks, service marks, or product names of the Licensor,
   except as required for reasonable and customary use in describing the
   origin of the Work and reproducing the content of the NOTICE file.

7. Disclaimer of Warranty. Unless required by applicable law or
   agreed to in writing, Licensor provides the Work (and each
   Contributor provides its Contributions) on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied, including, without limitation, any warranties or conditions
   of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
   PARTICULAR PURPOSE. You are solely responsible for determining the
   appropriateness of using or redistributing the Work and assume any
   risks associated with Your exercise of permissions under this License.

8. Limitation of Liability. In no event and under no legal theory,
   whether in tort (including negligence), contract, or otherwise,
   unless required by applicable law (such as deliberate and grossly
   negligent acts) or agreed to in writing, shall any Contributor be
   liable to You for damages, including any direct, indirect, special,
   incidental, or consequential damages of any character arising as a
   result of this License or out of the use or inability to use the
   Work (including but not limited to damages for loss of goodwill,
   work stoppage, computer failure or malfunction, or any and all
   other commercial damages or losses), even if such Contributor
   has been advised of the possibility of such damages.

9. Accepting Warranty or Additional Liability. While redistributing
   the Work or Derivative Works thereof, You may choose to offer,
   and charge a fee for, acceptance of support, warranty, indemnity,
   or other liability obligations and/or rights consistent with this
   License. However, in accepting such obligations, You may act only
   on Your own behalf and on Your sole responsibility, not on behalf
   of any other Contributor, and only if You agree to indemnify,
   defend, and hold each Contributor harmless for any liability
   incurred by, or claims asserted against, such Contributor by reason
   of your accepting any such warranty or additional liability.

END OF TERMS AND CONDITIONS

APPENDIX: How to apply the Apache License to your work.

   To apply the Apache License to your work, attach the following
   boilerplate notice, with the fields enclosed by brackets "[]"
   replaced with your own identifying information. (Don't include
   the brackets!)  The text should be enclosed in the appropriate
   comment syntax for the file format. We also recommend that a
   file or class name and description of purpose be included on the
   same "printed page" as the copyright notice for easier
   identification within third-party archives.

Copyright [yyyy] [name of copyright owner]

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""  # noqa

import ast
import inspect
import textwrap

from pharia_studio_sdk.evaluation.aggregation.aggregator import AggregationLogic
from pharia_studio_sdk.evaluation.evaluation.evaluator.evaluator import EvaluationLogic


class NotInteractiveEnvironmentError(Exception): ...


def is_running_interactively() -> bool:
    """Check if the code is running in an interactive environment."""
    try:
        from IPython import get_ipython

        return get_ipython() is not None
    except ModuleNotFoundError:
        return False


def get_notebook_source() -> str:
    """Get the source code of the running notebook."""
    from IPython import get_ipython

    shell = get_ipython()
    if shell is None:
        raise NotInteractiveEnvironmentError

    if not hasattr(shell, "user_ns"):
        raise AttributeError("Cannot access user namespace")

    # This is the list of input cells in the notebook
    in_list = shell.user_ns["In"]

    # Stitch them back into a single "file"
    full_source = "\n\n".join(cell for cell in in_list[1:] if cell)

    return full_source


def get_class_source(cls: type) -> str:
    """Get the latest source definition of a class in the notebook."""
    notebook_source = get_notebook_source()
    tree = ast.parse(notebook_source)
    class_name = cls.__name__

    # We need to walk the entire tree and get the last one since that's the most version of the cls
    segment = None
    for node in ast.walk(tree):
        if isinstance(node, ast.ClassDef) and node.name == class_name:
            segment = ast.get_source_segment(notebook_source, node)

    if segment is not None:
        return segment

    raise ValueError(f"Class '{class_name}' not found in the notebook")


def get_source_notebook_safe(logic: EvaluationLogic | AggregationLogic) -> str:
    # In ipython, we can't use inspect.getsource on classes defined in the notebook
    logic_class = type(logic)
    try:
        src = inspect.getsource(logic_class)
    except OSError:
        if is_running_interactively() and inspect.isclass(logic_class):
            src = get_class_source(logic_class)
    return textwrap.dedent(src)



from collections.abc import Iterable, Sequence
from typing import Optional, cast

from datasets import Dataset as HFDataset
from datasets import DatasetDict, IterableDataset, IterableDatasetDict
from pharia_inference_sdk.core.task import Input
from pydantic import BaseModel

from pharia_studio_sdk.connectors.base.json_serializable import SerializableDict
from pharia_studio_sdk.evaluation.dataset.dataset_repository import DatasetRepository
from pharia_studio_sdk.evaluation.dataset.domain import (
    Dataset,
    Example,
    ExpectedOutput,
)


class MultipleChoiceInput(BaseModel):
    question: str
    choices: Sequence[str]


class SingleHuggingfaceDatasetRepository(DatasetRepository):
    def __init__(
        self,
        huggingface_dataset: (
            DatasetDict | HFDataset | IterableDatasetDict | IterableDataset
        ),
    ) -> None:
        self._huggingface_dataset = huggingface_dataset

    def create_dataset(
        self,
        examples: Iterable[Example[Input, ExpectedOutput]],
        dataset_name: str,
        id: str | None = None,
        labels: set[str] | None = None,
        metadata: SerializableDict | None = None,
    ) -> Dataset:
        raise NotImplementedError

    def dataset(self, dataset_id: str) -> Dataset | None:
        raise NotImplementedError

    def dataset_ids(self) -> Iterable[str]:
        raise NotImplementedError

    def delete_dataset(self, dataset_id: str) -> None:
        raise NotImplementedError

    def example(
        self,
        dataset_id: str,
        example_id: str,
        input_type: type[Input],
        expected_output_type: type[ExpectedOutput],
    ) -> Example[Input, ExpectedOutput] | None:
        examples = self.examples(
            dataset_id=dataset_id,
            input_type=input_type,
            expected_output_type=expected_output_type,
        )

        for example in examples:
            if example.id == example_id:
                return example
        return None

    def examples(
        self,
        dataset_id: str,
        input_type: type[Input],
        expected_output_type: type[ExpectedOutput],
        examples_to_skip: Optional[frozenset[str]] = None,
    ) -> Iterable[Example[Input, ExpectedOutput]]:
        examples_to_skip = examples_to_skip or frozenset()
        answers = "ABCD"
        assert input_type == MultipleChoiceInput
        assert expected_output_type is str
        for index, sample in enumerate(self._huggingface_dataset["test"]):
            if str(index) not in examples_to_skip:
                yield Example(
                    input=cast(
                        Input,
                        MultipleChoiceInput(
                            question=sample["question"], choices=sample["choices"]
                        ),
                    ),
                    expected_output=cast(ExpectedOutput, answers[sample["answer"]]),
                    id=str(index),
                )



from typing import Generic, Optional, TypeVar
from uuid import uuid4

from pharia_inference_sdk.core.task import Input
from pharia_inference_sdk.core.tracer.tracer import PydanticSerializable
from pydantic import BaseModel, Field
from rich.tree import Tree

from pharia_studio_sdk.connectors.base.json_serializable import (
    SerializableDict,
)

ExpectedOutput = TypeVar("ExpectedOutput", bound=PydanticSerializable)
"""Dataset-specific type that defines characteristics that an :class:`Output` can be checked against.

Traditional names for this are `label` or `y` in classification."""


class Example(BaseModel, Generic[Input, ExpectedOutput]):
    """Example case used for evaluations.

    Attributes:
        input: Input for the :class:`Task`. Has to be same type as the input for the task used.
        expected_output: The expected output from a given example run.
            This will be used by the evaluator to compare the received output with.
        id: Identifier for the example, defaults to uuid.
        metadata: Optional dictionary of custom key-value pairs.

    Generics:
        Input: Interface to be passed to the :class:`Task` that shall be evaluated.
        ExpectedOutput: Output that is expected from the run with the supplied input.
    """

    input: Input
    expected_output: ExpectedOutput
    id: str = Field(default_factory=lambda: str(uuid4()))
    metadata: Optional[SerializableDict] = None

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return (
            f"Example ID = {self.id}\n"
            f"Input = {self.input}\n"
            f"Expected output = {self.expected_output}\n"
            f"Metadata = {self.metadata}\n"
        )

    def _rich_render(self) -> Tree:
        example_tree = Tree(f"Example: {self.id}")
        example_tree.add("Input").add(str(self.input))
        example_tree.add("Expected Output").add(str(self.expected_output))
        if self.metadata:
            example_tree.add("Metadata").add(str(self.metadata))
        return example_tree


class Dataset(BaseModel):
    """Represents a dataset linked to multiple examples.

    Attributes:
        id: Dataset ID.
        name: A short name of the dataset.
        label: Labels for filtering datasets. Defaults to empty list.
        metadata: Additional information about the dataset. Defaults to empty dict.
    """

    id: str = Field(default_factory=lambda: str(uuid4()))
    name: str
    labels: set[str] = set()
    metadata: SerializableDict = dict()

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return (
            f"Dataset ID = {self.id}\n"
            f"Name = {self.name}\n"
            f"Labels = {self.labels}\n"
            f"Metadata = {self.metadata}"
        )



import warnings
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Optional

from fsspec import AbstractFileSystem  # type: ignore
from fsspec.implementations.local import LocalFileSystem  # type: ignore
from pharia_inference_sdk.core import FileTracer, InMemoryTracer, JsonSerializer, Output
from pharia_inference_sdk.core.tracer.tracer import Tracer

from pharia_studio_sdk.evaluation.infrastructure.file_system_based_repository import (
    FileSystemBasedRepository,
)
from pharia_studio_sdk.evaluation.run.domain import (
    ExampleOutput,
    FailedExampleRun,
    RunOverview,
)
from pharia_studio_sdk.evaluation.run.run_repository import RecoveryData, RunRepository


class FileSystemRunRepository(RunRepository, FileSystemBasedRepository):
    TMP_FILE_TYPE: str = "tmp"

    def __init__(self, file_system: AbstractFileSystem, root_directory: Path) -> None:
        FileSystemBasedRepository.__init__(self, file_system, root_directory)
        RunRepository.__init__(self)

    def store_run_overview(self, overview: RunOverview) -> None:
        self.write_utf8(
            self._run_overview_path(overview.id),
            overview.model_dump_json(indent=2),
            create_parents=True,
        )
        # create empty folder just in case no examples are ever saved
        self.mkdir(self._run_directory(overview.id))

    def _tmp_file_path(self, tmp_hash: str) -> Path:
        return self._run_directory(tmp_hash + "." + self.TMP_FILE_TYPE)

    def _create_temporary_run_data(self, tmp_hash: str, run_id: str) -> None:
        self.write_utf8(
            self._tmp_file_path(tmp_hash),
            RecoveryData(run_id=run_id).model_dump_json(),
            create_parents=True,
        )

    def _delete_temporary_run_data(self, tmp_hash: str) -> None:
        self.remove_file(self._tmp_file_path(tmp_hash))

    def _temp_store_finished_example(self, tmp_hash: str, example_id: str) -> None:
        data = RecoveryData.model_validate_json(
            self.read_utf8(self._tmp_file_path(tmp_hash))
        )
        data.finished_examples.append(example_id)
        self.write_utf8(
            self._tmp_file_path(tmp_hash),
            data.model_dump_json(),
            create_parents=True,
        )

    def finished_examples(self, tmp_hash: str) -> Optional[RecoveryData]:
        try:
            return RecoveryData.model_validate_json(
                self.read_utf8(self._tmp_file_path(tmp_hash))
            )
        except FileNotFoundError:
            return None

    def run_overview(self, run_id: str) -> Optional[RunOverview]:
        file_path = self._run_overview_path(run_id)
        if not self.exists(file_path):
            return None

        content = self.read_utf8(file_path)
        return RunOverview.model_validate_json(content)

    def run_overview_ids(self) -> Sequence[str]:
        return sorted(self.file_names(self._run_root_directory()))

    def store_example_output(self, example_output: ExampleOutput[Output]) -> None:
        serialized_result = JsonSerializer(root=example_output)
        self.write_utf8(
            self._example_output_path(example_output.run_id, example_output.example_id),
            serialized_result.model_dump_json(indent=2),
            create_parents=True,
        )

    def example_output(
        self, run_id: str, example_id: str, output_type: type[Output]
    ) -> Optional[ExampleOutput[Output] | ExampleOutput[FailedExampleRun]]:
        file_path = self._example_output_path(run_id, example_id)
        if not self.exists(file_path):
            warnings.warn(
                f'Repository does not contain a run with id: "{run_id}"', UserWarning
            )
            return None
        content = self.read_utf8(file_path)
        # mypy does not accept dynamic types
        return ExampleOutput[output_type].model_validate_json(  # type: ignore
            json_data=content
        )

    def example_outputs(
        self, run_id: str, output_type: type[Output]
    ) -> Iterable[ExampleOutput[Output] | ExampleOutput[FailedExampleRun]]:
        path = self._run_output_directory(run_id)
        if not self.exists(path):
            warnings.warn(
                f'Repository does not contain a run with id: "{run_id}"', UserWarning
            )
            return []

        example_outputs = []
        for file_name in self.file_names(path):
            output = self.example_output(run_id, file_name, output_type)
            if output is not None:
                example_outputs.append(output)

        return sorted(
            example_outputs,
            key=lambda _example_output: _example_output.example_id,
        )

    def example_output_ids(self, run_id: str) -> Sequence[str]:
        return sorted(self.file_names(self._run_output_directory(run_id)))

    def _run_root_directory(self) -> Path:
        path = self._root_directory / "runs"
        return path

    def _run_directory(self, run_id: str) -> Path:
        path = self._run_root_directory() / run_id
        return path

    def _run_output_directory(self, run_id: str) -> Path:
        path = self._run_directory(run_id) / "output"
        return path

    def _run_overview_path(self, run_id: str) -> Path:
        return self._run_directory(run_id).with_suffix(".json")

    def example_tracer(self, run_id: str, example_id: str) -> Optional[Tracer]:
        file_path = self._example_trace_path(run_id, example_id)
        if not self.exists(file_path):
            return None
        return self._parse_log(file_path)

    def create_tracer_for_example(self, run_id: str, example_id: str) -> Tracer:
        file_path = self._example_trace_path(run_id, example_id)
        return FileTracer(file_path)

    def _trace_directory(self, run_id: str) -> Path:
        path = self._run_directory(run_id) / "trace"
        return path

    def _example_trace_path(self, run_id: str, example_id: str) -> Path:
        return (self._trace_directory(run_id) / example_id).with_suffix(".jsonl")

    @staticmethod
    def _parse_log(log_path: Path) -> InMemoryTracer:
        return FileTracer(log_path).traces()

    def _example_output_path(self, run_id: str, example_id: str) -> Path:
        return (self._run_output_directory(run_id) / example_id).with_suffix(".json")


class FileRunRepository(FileSystemRunRepository):
    def __init__(self, root_directory: Path) -> None:
        super().__init__(LocalFileSystem(), root_directory)

    @staticmethod
    def path_to_str(path: Path) -> str:
        return str(path)



from .aggregation.accumulator import MeanAccumulator as MeanAccumulator
from .aggregation.aggregation_repository import (
    AggregationRepository as AggregationRepository,
)
from .aggregation.aggregator import AggregationLogic as AggregationLogic
from .aggregation.aggregator import Aggregator as Aggregator
from .aggregation.domain import AggregatedEvaluation as AggregatedEvaluation
from .aggregation.domain import AggregationOverview as AggregationOverview
from .aggregation.elo_aggregation import AggregatedComparison as AggregatedComparison
from .aggregation.elo_aggregation import (
    ComparisonEvaluationAggregationLogic as ComparisonEvaluationAggregationLogic,
)
from .aggregation.elo_aggregation import EloCalculator as EloCalculator
from .aggregation.elo_aggregation import (
    MatchesAggregationLogic as MatchesAggregationLogic,
)
from .aggregation.elo_aggregation import WinRateCalculator as WinRateCalculator
from .aggregation.file_aggregation_repository import (
    FileAggregationRepository as FileAggregationRepository,
)
from .aggregation.hugging_face_aggregation_repository import (
    HuggingFaceAggregationRepository as HuggingFaceAggregationRepository,
)
from .aggregation.in_memory_aggregation_repository import (
    InMemoryAggregationRepository as InMemoryAggregationRepository,
)
from .benchmark.studio_benchmark import StudioBenchmark as StudioBenchmark
from .benchmark.studio_benchmark import (
    StudioBenchmarkRepository as StudioBenchmarkRepository,
)
from .dataset.dataset_repository import DatasetRepository as DatasetRepository
from .dataset.domain import Dataset as Dataset
from .dataset.domain import Example as Example
from .dataset.domain import ExpectedOutput as ExpectedOutput
from .dataset.file_dataset_repository import (
    FileDatasetRepository as FileDatasetRepository,
)
from .dataset.hugging_face_dataset_repository import (
    HuggingFaceDatasetRepository as HuggingFaceDatasetRepository,
)
from .dataset.in_memory_dataset_repository import (
    InMemoryDatasetRepository as InMemoryDatasetRepository,
)
from .dataset.single_huggingface_dataset_repository import (
    MultipleChoiceInput as MultipleChoiceInput,
)
from .dataset.single_huggingface_dataset_repository import (
    SingleHuggingfaceDatasetRepository as SingleHuggingfaceDatasetRepository,
)
from .dataset.studio_dataset_repository import (
    StudioDatasetRepository as StudioDatasetRepository,
)
from .evaluation.domain import Evaluation as Evaluation
from .evaluation.domain import EvaluationFailed as EvaluationFailed
from .evaluation.domain import EvaluationOverview as EvaluationOverview
from .evaluation.domain import ExampleEvaluation as ExampleEvaluation
from .evaluation.domain import FailedExampleEvaluation as FailedExampleEvaluation
from .evaluation.evaluation_repository import (
    EvaluationRepository as EvaluationRepository,
)
from .evaluation.evaluator.argilla_evaluator import (
    ArgillaEvaluationLogic as ArgillaEvaluationLogic,
)
from .evaluation.evaluator.argilla_evaluator import ArgillaEvaluator as ArgillaEvaluator
from .evaluation.evaluator.argilla_evaluator import (
    InstructComparisonArgillaEvaluationLogic as InstructComparisonArgillaEvaluationLogic,
)
from .evaluation.evaluator.argilla_evaluator import (
    RecordDataSequence as RecordDataSequence,
)
from .evaluation.evaluator.async_evaluator import (
    AsyncEvaluationRepository as AsyncEvaluationRepository,
)
from .evaluation.evaluator.evaluator import EvaluationLogic as EvaluationLogic
from .evaluation.evaluator.evaluator import Evaluator as Evaluator
from .evaluation.evaluator.evaluator import (
    SingleOutputEvaluationLogic as SingleOutputEvaluationLogic,
)
from .evaluation.evaluator.incremental_evaluator import (
    ComparisonEvaluation as ComparisonEvaluation,
)
from .evaluation.evaluator.incremental_evaluator import (
    EloEvaluationLogic as EloEvaluationLogic,
)
from .evaluation.evaluator.incremental_evaluator import (
    EloGradingInput as EloGradingInput,
)
from .evaluation.evaluator.incremental_evaluator import (
    IncrementalEvaluationLogic as IncrementalEvaluationLogic,
)
from .evaluation.evaluator.incremental_evaluator import (
    IncrementalEvaluator as IncrementalEvaluator,
)
from .evaluation.evaluator.incremental_evaluator import Matches as Matches
from .evaluation.evaluator.incremental_evaluator import MatchOutcome as MatchOutcome
from .evaluation.file_evaluation_repository import (
    AsyncFileEvaluationRepository as AsyncFileEvaluationRepository,
)
from .evaluation.file_evaluation_repository import (
    FileEvaluationRepository as FileEvaluationRepository,
)
from .evaluation.graders import BleuGrader as BleuGrader
from .evaluation.graders import FScores as FScores
from .evaluation.graders import HighlightCoverageGrader as HighlightCoverageGrader
from .evaluation.graders import LanguageMatchesGrader as LanguageMatchesGrader
from .evaluation.graders import RougeGrader as RougeGrader
from .evaluation.in_memory_evaluation_repository import (
    AsyncInMemoryEvaluationRepository as AsyncInMemoryEvaluationRepository,
)
from .evaluation.in_memory_evaluation_repository import (
    InMemoryEvaluationRepository as InMemoryEvaluationRepository,
)
from .infrastructure.hugging_face_repository import (
    HuggingFaceRepository as HuggingFaceRepository,
)
from .infrastructure.repository_navigator import (
    RepositoryNavigator as RepositoryNavigator,
)
from .infrastructure.repository_navigator import (
    aggregation_overviews_to_pandas as aggregation_overviews_to_pandas,
)
from .infrastructure.repository_navigator import (
    evaluation_lineages_to_pandas as evaluation_lineages_to_pandas,
)
from .infrastructure.repository_navigator import (
    run_lineages_to_pandas as run_lineages_to_pandas,
)
from .run.domain import ExampleOutput as ExampleOutput
from .run.domain import RunOverview as RunOverview
from .run.domain import SuccessfulExampleOutput as SuccessfulExampleOutput
from .run.file_run_repository import FileRunRepository as FileRunRepository
from .run.in_memory_run_repository import InMemoryRunRepository as InMemoryRunRepository
from .run.run_repository import RunRepository as RunRepository
from .run.runner import Runner as Runner

__all__ = [symbol for symbol in dir()]
